/******************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:     orva_DiagPlot.sas                                             
 * PURPOSE:  Generate data to plot diagnostic graphs               
 * SUPPORT:  jariou         
 *
 * Parameters:
 *
 *   data      = Specifies the name of the loss data set to be analyzed.
 *               Must contain the following columns:
 *                 AMOUNT
 *                 THRESHOLD
 *                 FINANCIALDATE
 *   ccUid     = key for this cell.
 *   dist      = Specifies the name of the distribution.
 *   parmdata  = Name of the input data set containing parameter estimates.
 *   maxPoints =
 *   threshold
 *   defaultThreshold =
 *   diagdata  = The name of the output dataset.
 *               It contains the following variables:
 *                - AMOUNT  :   the loss amount
 *                - FINANCIALDATE : date of loss
 *                - QUANT     :   the corresponding model quantile
 *                - FREQ      :   the model density function value
 *                - TCDF      :   the model cumulative distribution function value
 *                - ECDF      :   the empirical distribution function value
 *                - CDFDIF    :   the difference between model and empirical cdf
 *
 ******************************************************************************/
%macro orva_DiagPlot(
		data=,      
	    ccUid=,   
        dist=, 
        parmdata=,  
		maxPoints=200,
        threshold=,   
	    defaultThreshold=0,
	    diagdata=,
	    plotStatus=plotStatus   
        );

        %if &defaultThreshold=%str() %then %do;
		%let defaultThreshold=0.000000001;
	%end;


	data _workdata_;
		set &data; where CCUID=&ccUid;
		if THRESHOLD=. then THRESHOLD = &defaultThreshold;
	run;
	/*- If threshold is passed as parameter,
		then put it in the dataset as a repeated value	-*/
	%if &threshold ^= %str() %then %do;
		%if %sysevalf(&threshold<0)  %then %do;
			%put ERROR: THRESHOLD must be non-negative.;
			%let abort=1;
			%let status=1;      
			%let reason=NEGATIVE THRESHOLD;
			%goto DONE; 
		%end;       

		data _workdata_;
			set _workdata_;
			THRESHOLD=&threshold;
		run;

		%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
			%put ERROR: Unable to set threshold value.;
			%let abort=1;
			%let reason=Unable to set threshold value;
			%goto DONE;
		%end;

	%end;

	/*** set up the status macro variables	***/
	%global status reason;
	%let status=0;
	%let reason=Successful;
	%let dist=%upcase(&dist);
	
	/*** force load of distribution macro ***/
	%orvd_&dist
	%let abort=0;
	%let BinMax=.;

	/*** create the status data set upfront **/
	data &plotStatus;
		format status 8. reason $100.;
		status=1;
		reason="Unknown";
		count=0;
	run;

	/*** keep only the parm data corresponding to the distribution given ***/
	data _parm_;
		set &parmdata(where=(distribution="&dist"));
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
		%put ERROR: Unable to create parameter data set.;
		%let abort=1;
        %let reason=Unable to create parameter data set;
		%goto DONE;
	%end;

	/*** Make sure the parm dataset is not empty ***/
	%let dsid=%sysfunc(open(_parm_,I));
	%let nparmobs=%sysfunc(attrn(&dsid,nlobs));
	%let rc=%sysfunc(close(&dsid));
    %if %sysevalf(&nparmobs=0, boolean) %then %do;
		%put ERROR: Unable to find parms in the parmdata.;
		%let abort=1;
		%let reason=Unable to find parms in the parmdata;
		%goto DONE;
	%end;

	/*** transpose the parameters dataset ***/
	proc transpose data=_parm_ out=_parm_; 
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
		%put ERROR: Unable to transpose parameter data set.;
		%let abort=1;
		%let reason=Unable to transpose parameter data set;
		%goto DONE;
	%end;



	/*** Assign corresponding values to macros: %location, %scale, %shape1, %shape2 ***/
	data _null_;
		set _parm_;
           if upcase(_name_) ne "CCUID" and upcase(_name_) ne "STATUS" and upcase(_name_) ne "REASON" then 
		   call symput(_name_, col1);
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to create parameter macros.;
	    %let abort=1;
	    %let reason=Unable to create parameter macros;
	    %goto DONE;
	%end;

        /* Adjust amounts with the KRIS ------------------------*/
        /* See if there are KRIs */
        data _null_;
	  set &parmdata;
			
	  call symput("KRI_ADJUST",compress(COUNT_KRI));
        run;	

    %if %sysevalf(&KRI_ADJUST >0, boolean) %then %do;
	    %orva_get_kri_log(data=&parmdata);
    %end;
    %else %do;
        %let KRI_LOG=0; 
	    %let KRI_LOG_AVG=0; 
    %end;

    %if %sysevalf(&KRI_ADJUST > 0, boolean) %then %do ;
            data _temp_;
		        set _workdata_;
		        where  AMOUNT > THRESHOLD;

		        nonadjusted_AMOUNT    = AMOUNT;
		        nonadjusted_THRESHOLD = THRESHOLD;		        
                log_kri               = exp( &KRI_LOG );
                THRESHOLD             = THRESHOLD / log_kri;
                AMOUNT                = AMOUNT / log_kri; 

                keep nonadjusted_AMOUNT nonadjusted_THRESHOLD AMOUNT THRESHOLD FINANCIALDATE CCUID;
	        run;

        %let Param_1              = &Orig_Param_1;
        %end;
        %else %do;
			/*** subset the data only for values where AMOUNT > THRESHOLD ***/
			data _temp_;
				set _workdata_;
				where AMOUNT > THRESHOLD;
		        nonadjusted_AMOUNT = AMOUNT;
		        nonadjusted_THRESHOLD = THRESHOLD;
		        keep nonadjusted_AMOUNT nonadjusted_THRESHOLD AMOUNT THRESHOLD FINANCIALDATE CCUID;
			run;
		%end;

	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to subset the loss data.;
	    %let abort=1;
	    %let reason=Unable to subset the loss data;
	    %goto DONE;
	%end;

        
	/*** sort the data by loss amount ***/
	proc sort data=_temp_;
		by AMOUNT;
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to sort loss data.;
	    %let abort=1;
	    %let reason=Unable to sort loss data;
	    %goto DONE;
	%end;

	/*- Find the smallest threshold point -*/
	/*- Put the value in mintreshold macro -*/
	data _null_;
		set _temp_ nobs=nobs;
		retain minthreshold 1e16;
		minthreshold=min(minthreshold,THRESHOLD);
		if _N_=nobs then do;
			call symput("minthreshold",minthreshold);
		end;
	run;

	/*** sample data (reduce extra points) if requested ***/
	%if &maxPoints ne %str() %then %do;
	    %FlushExtraPoints(data=_temp_,max=&maxPoints)
	%end;

    /*--- Set all the Threshold to the smallest one ---*/
    data _temp_;
        set _temp_ nobs=nobs;
        ecdf=_N_/(nobs+1);
        THRESHOLD = &minThreshold;
    run;
	

	%put          *************************;
	%put          Distribution=      &dist;
	%put          *************************;

	/*- do the actual QQ-Plot calculation -*/
	/*- delegate to macros for individual distributions -*/
	/*- for some odd reason, this macro name needs to be assembled before it is called -*/
	/*	ReeU */
		%let QuantMacro=Quant&dist;
		%&QuantMacro
	/*	ReeU */

	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
		%put ERROR: Unable to compute quantiles.;
		%let abort=1;
		%let reason=Unable to compute quantiles;
		%goto DONE;
	%end;

	/*- create the data for fitted curve -*/
	data _temp_;
        set _temp_ nobs=obsCount;
		retain maxfreq 0;
		/* ReeU */
	    %Pdf&dist(AMOUNT,freq)
	    %Cdf&dist(THRESHOLD,truncdf)
		%Cdf&dist(AMOUNT, tcdf)
		/* ReeU */
		freq=freq/(1-truncdf);   
		maxfreq=max(freq,maxfreq);

        ecdf=_N_/(obsCount+1);
		tcdf=(tcdf-truncdf)/(1-truncdf);
		cdfdif=tcdf-ecdf;

		if _N_=1 then do;
			call symput("minLoss",AMOUNT);
		end;
        if _N_=obsCount then do;
		   call symput("maxfreq",maxfreq);
		   call symput("count",_N_);
		   call symput("maxLoss",AMOUNT);
		   call symput("BinMax",AMOUNT);
		end;
	run;

	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to create fitted curve data set.;
	    %let abort=1;
	    %let reason=Unable to create fitted curve data set;
	    %goto DONE;
	%end;

	%let BinMax=&maxLoss; /* just to make sure it is well defined at this point */

    /*- find points in right tail with freq < maxfreq/50   -*/
	/*- this sets a default area to inspect the data in    -*/
	data _null_;
        set _temp_ nobs=obsCount;
		retain cutoff 0;

        if _N_>obsCount / 2 and freq < (&maxfreq)/50 then do;
			cutoff=1;
		end;
        else do;
			cutoff=0;
		end;

		if cutoff=1 and lag(cutoff)=0 then do;
			call symput("BinMax",AMOUNT);
		end;
	run;
 

	/*	This part constructs a dataset to plot the density starting at the minimum threshold point*/
	/*	We simply add 20 points between the minimum threshold and the minimum loss so we can get a good*/
	/*	looking curve in that region*/
/*    %let stepSize = (&minLoss - &minThreshold)/20;*/
/**/
/*    data _temp1_;*/
/*        stepSize=&stepSize;*/
/*        do i=0 to 19;*/
/*			AMOUNT = &minThreshold + i*stepSize;*/
/*			THRESHOLD=&minThreshold;*/
/*		    %Cdf&dist(THRESHOLD,truncdf)*/
/*		    %Pdf&dist(AMOUNT,freq)*/
/*            freq=freq/(1-truncdf);*/
/*            output;*/
/*        end;*/
/*    run;*/

	
	/*** write resultant dataset ***/
	data &diagdata(keep= AMOUNT quant tcdf ecdf cdfdif freq FINANCIALDATE THRESHOLD nonadjusted_AMOUNT nonadjusted_THRESHOLD);
        set _temp_;
    run;

/*	data &diagdata.Density(keep= AMOUNT freq);*/
/*            set _temp1_ _temp_;*/
/*    run;*/
	
%DONE:

	%if &abort %then %do;
	    %put ERROR: orva_diagplot macro ended abnormally.;
	    %let status=1;
	%end;

	/*- delete unwanted data -*/
	proc datasets lib=work memtype=data nolist;
		delete _parm_  _stat_ _temp_ _temp1_  ;
	run;
	quit;

	data &plotStatus;
		format status 8. reason $100.;
		status=&status;
		reason="&reason";
		count=&count;
		BinMin=&minthreshold;
		BinMax=&BinMax;
		LossMax=&maxLoss;
	run;
%mend orva_Diagplot;


%macro FlushExtraPoints(data= , max= );
	data &data(drop=nn count);
		retain nn count;
		set &data nobs=nobs;

		if _N_= 1 then do;
			count=round(nobs/&max);
			if count=0 then count=1;
			nn = count - 1 ;
		end;

		nn = nn+1;

		if nn ne count  then do;
           if _N_ ne nobs then do;
		   delete;
		end;
        end;
		else do;
		   nn=0;
		end;
	run;

%mend FlushExtraPoints;
