/*****************************************************************************************
 * Copyright (c) 2004 by SAS Institute Inc.  Cary NC 27513  USA
 *
 * NAME: ORVA_freq_kri_effect.sas                                             
 * DESCRIPTION: Compute the effect of the kris's on the frequency distribution
 * SUPPORT: jariou       
 *
 * PARAMETERS: 
 *
 *    freqPriorityDataset = Name of the frequency priority dataset. 
 *                          The data set has one row for each cell containing:
 *                              SOURCE_TYPE
 *                              SOURCE_NAME
 *                              DATASET_NAME
 *                              CCUID                        
 *    freqKriDatasetBl    = Dataset with the simulation values for the Businessline
 *                          frequency KRI variables. Must have:
 *                             BLUID_KRI and <risk_factors>
 *    freqFriDatasetFirm  = Dataset with the simulation values for the Firm
 *                          frequency KRI variables. Must have:
 *                             BLUID_KRI and <risk_factors>
 *    ccData              = Name of the Cross Classification dataset containing:
 *                              CCUID, CCVUID_0, CCVUID_1
 *    outStatus           = Success or failure, default is outstatus 
 *    outFParmEffect      = Frequency parameters output dataset               
 *    blEstimates         = Dataset created by %orva_freq_fit_kri() containing
 *                          parameter estimates for the businessline dummy variables.  
 *                          This dataset will be modified by this macro!
 *                                               
 ****************************************************************/
%macro ORVA_freq_kri_effect(freqPriorityDataset = ,  
                            freqKriDatasetBl    = ,
                            freqKriDatasetFirm  = ,
                            ccData              = ,
                            outStatus           = outStatus,
                            outFParmEffect      = outfParmEffect,
                            blEstimates         =
                            );
/*%orvh_displaypaths;*/
	%orva_dist_globals;

    %let status=0;
    %let reason=Successful;
    %let abort=0;

    /*- create the status data set upfront -*/
    data &outstatus;
        format status 8. reason $100.;
        status=1;
        reason="Unknown";
    run;
 
    %if "&freqPriorityDataset" = "" %then %do;
        %put ERROR: The parameter freqPriorityDataset= must be specified.;
        %let abort=1;
        %let reason=MISSING freqPriorityDataset= PARAMETER;
        %goto DONE;
    %end;


	%let ds_nobs=0;
	%let dsid=%sysfunc(open(&freqPriorityDataset,I));
	%if &dsid NE 0 %then %do;
		%let ds_nobs=%sysfunc(attrn(&dsid,nlobs));
		%let rc=%sysfunc(close(&dsid));
	%end;
	%else %do;
        %put ERROR: Cannot open file freqPriorityDataset=.;
        %let abort=1;
        %let reason=CANNOT OPEN freqPriorityDataset= PARAMETER;
        %goto DONE;		
	%end;
	%if &ds_nobs LE 0 %then %do; /* should have at least one obs */
	    %put ERROR: The dataset &freqPriorityDataset is empty (has no observations).;
        %let abort=1;
        %let reason=The dataset &freqPriorityDataset is empty (has no observations);
        %goto DONE;
	%end;


    /***** get KRI dataset *****/
	%let _freqKriAll_exists_ = 1;
	%let _only_Firm_KRIs_ = 0;
    %if "&freqKriDatasetBl" ne "" and "&freqKriDatasetFirm" ne "" %then %do;   
		/* This merge works correctly:  the variable values in the single obs found in fKDFirm are added 
		as a new column to every obs in fKDBL. */
	    data _freqKriAll_ ;
	        merge &freqKriDatasetBl &freqKriDatasetFirm;
	        by risk_factor_variable_date;
	        rename BLUID_KRI=BLUID;
	    run;
        proc sort data= _freqKriAll_; by BLUID; run;
    %end;
    %else %if "&freqKriDatasetBl" ne "" %then %do;
	    data _freqKriAll_;
	        set &freqKriDatasetBl;
	        rename BLUID_KRI=BLUID;
	    run;
       proc sort data= _freqKriAll_; by BLUID; run;
    %end;
    %else %if "&freqKriDatasetFirm" ne "" %then %do;
		%let _only_Firm_KRIs_ = 1;
	    data _freqKriAll_;
	        set &freqKriDatasetFirm;
/*			BLUID = .  Stacey added this, Laura removed it -- it messes up my set statements later on... */  
	    run;
    %end;  
	%else %let _freqKriAll_exists_ = 0;
    
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to merge all KRIs together.;
        %let abort=1;
        %let reason=Unable to merge all KRIs together;
        %goto DONE;
    %end;
        
	
	data __tmp__;
		set &freqPriorityDataset ;
		where SOURCE_TYPE NE "R";
		call symputX("NROWS", _N_);
	run;

    /* Find the number of rows in freqPriorityDataset */
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to compute the number of rows in the simulation priority dataset.;
        %let abort=1;
        %let reason=Unable to compute the number of rows in the simulation priority dataset;
        %goto DONE;
    %end;


    /*    Get the list of CCUID's for which calculations need to be made*/
    proc SQL noprint;
            select 
                    dataset_name, 
                    CCUID, 
                    SOURCE_TYPE, 
                    SCENARIO_TYPE 
            into 
                    :DSNAME1 - :DSNAME&NROWS , 
                    :CCUID1 - :CCUID&NROWS , 
                    :SOURCE_TYPE1 - :SOURCE_TYPE&NROWS, 
                    :SCENARIO_TYPE1 - :SCENARIO_TYPE&NROWS 
            from 
                    __tmp__ 
            ;
    quit;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to create the macros containing dataset_name(s), CCUID(s), and Source_Type(s).;
        %let abort=1;
        %let reason=Unable to create the macros containing dataset_name(s), CCUID(s), and Source_Type(s);
        %goto DONE;
    %end;


    /* Loop through the rows to calculate the KRI & BL effects */
    %do i=1 %to &NROWS;

		%if "&&&SOURCE_TYPE&i" NE "I" %then %do;
			%if &i eq 1 %then %do;
                data _foo_; 
					length Distribution $12;
                    set &DSNAME1;
                    where CCUID=&CCUID1;
                    %if "&&&SOURCE_TYPE&i" eq "S" and 
                        "&&&SCENARIO_TYPE&i" eq "B" %then %do;
                        if Fparam_2 = . then Distribution = "POIS   ";
                        else Distribution = "NEGBIN  ";
                    %end;
                run;
            %end;
            %else %do;
	            data _TMP_; 
					length Distribution $12;
	                set &&&DSNAME&i;
	                where CCUID=&&&CCUID&i;
	                %if "&&&SOURCE_TYPE&i" eq "S" and 
	                        "&&&SCENARIO_TYPE&i" eq "B" %then %do;
                                if Fparam_2 = . then Distribution = "POIS   ";
                                else Distribution = "NEGBIN  ";
	                %end;
	            run;
	            data _foo_;
	                set _foo_ _tmp_;
	            run; 
			%end;
			%goto NEXTCCUID;  /* We are done for this row.  Now do the next row (in the big do-loop).  */
		%end;   /* if "&&&SOURCE_TYPE&i" NE "I"  */


		/* ************************************************************************************************ */
		/* ********** The rest of this do-loop is only for INTERNAL data (i.e., source_type == I)********** */

		/* Create macro COUNT_KRI = # of KRI's used */
        
        %let COUNT_KRI = -1;        /* initializing */
		%let FREQ_ITERATION = -1;  /* initializing */
        /* Consortium data has no KRIs, so COUNT_KRI will be 0; Internal data may or may not have KRIs */

        PROC SQL noprint;
            select COUNT_KRI, Iteration into :COUNT_KRI, :FREQ_ITERATION
            from &&&DSNAME&i
            where CCUID=&&&CCUID&i;
        QUIT;
        %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
            %put ERROR: Unable to obtain the number of selected KRI(S) variables.;
            %let abort=1;
            %let reason=Unable to obtain the number of selected KRI(S) variables;
            %goto DONE;
        %end;
		%if %sysevalf(&COUNT_KRI+0 LE 0, BOOLEAN) %then %do;  /* No KRIs were selected, so Fparams will remain unchanged */

/* less good: %sysevalf(&COUNT_KRI+0, BOOLEAN) LE 0*/

			%if &i eq 1 %then %do;
                data _foo_; 
					length Distribution $12;
                    set &DSNAME1;
                    where CCUID=&CCUID1;
				run;
            %end;
            %else %do;
	            data _TMP_; 
					length Distribution $12;
	                set &&&DSNAME&i;
	                where CCUID=&&&CCUID&i;
	            run;
	            data _foo_;
	                set _foo_ _tmp_;
	            run; 
			%end;
			%goto NEXTCCUID;  /* We are done for this row.  Now do the next row (in the big do-loop).  */
		%end;

		/* ************************************************************************************************ */
		/* ********** The rest of this do-loop is only for INTERNAL data WITH KRIs *********************** */
	
		%let FREQ_KRI_LIST = ;  /* initializing */

        PROC SQL noprint;
            select SelectedKRIs into :FREQ_KRI_LIST
            from &&&DSNAME&i
				/* Fix (1/2) for wrong way of keeping fits --BEGIN-- */
            where CCUID=&&&CCUID&i AND status=0; /* Only after the AND */
				/* Fix (1/2) for wrong way of keeping fits --BEGIN-- */
        QUIT;
        %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
            %put ERROR: Unable to obtain the list of selected KRI(S) variables.;
            %let abort=1;
            %let reason=Unable to obtain the list of selected KRI(S) variables;
            %goto DONE;
        %end;

		%let KEPT_KRI_BLUID_COUNT = -1; /* initializing */
 		%let ALL_KRI_BLUID_COUNT = -1;  /* initializing */

        /* Create KEPT_KRI_BLUID_COUNT = # of significant business lines panels w/ this CCUID */
		/*    and ALL_KRI_BLUID_COUNT = # of all possible business lines panels w/ this CCUID */
		%let ds_nobs=0;
		%let dsid=%sysfunc(open(&blEstimates,I));
		%if &dsid NE 0 %then %do;
			%let ds_nobs=%sysfunc(attrn(&dsid,nlobs));
			%let rc=%sysfunc(close(&dsid));
		%end;
		%else %do;  /* When could this possibly happen? Print error msg to log, but end gracefully for this CCUID */
			%put ERROR: The open for dataset blEstimates=&blEstimates failed; CCUID=&&&CCUID&i;
			%if &i eq 1 %then %do;
                data _foo_; 
					length Distribution $12;
                    set &DSNAME1;
                    where CCUID=&CCUID1;
				run;
            %end;
            %else %do;
	            data _TMP_; 
					length Distribution $12;
	                set &&&DSNAME&i;
	                where CCUID=&&&CCUID&i;
	            run;
	            data _foo_;
	                set _foo_ _tmp_;
	            run; 
			%end;
			%goto NEXTCCUID;  /* We are done for this row.  Now do the next row (in the big do-loop).  */
        %end;
		%if &ds_nobs LE 0 %then %do; /* should have at least one obs */
		    %put ERROR: The dataset &blEstimates is empty (has no observations).;
            %let abort=1;
            %let reason=The dataset &blEstimates is empty (has no observations);
            %goto DONE;
		%end;
		%else %do;
            proc SQL noprint;
                select 
                        count(BLUID) 
                into    
                        :KEPT_KRI_BLUID_COUNT
                from 
                        &blEstimates
                where 
                        Estimate NE .
                    AND
                        Iteration = &FREQ_ITERATION;
            quit;
            %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
                %put ERROR: Unable to count number of significant business lines used as panels in count regression.;
                %let abort=1;
                %let reason=Unable to count number of significant business lines used as panels in count regression;
                %goto DONE;
            %end;
			proc SQL noprint;
                select 
                        count(BLUID) 
                into    
                        :ALL_KRI_BLUID_COUNT
                from 
                        &blEstimates
                where 
                        Iteration = &FREQ_ITERATION;
            quit;
            %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
                %put ERROR: Unable to count number of all business lines in this cell.;
                %let abort=1;
                %let reason=Unable to count number of all business lines in this cell;
                %goto DONE;
            %end;
        %end;


        %if &KEPT_KRI_BLUID_COUNT > 0 %then %do;
			proc SQL noprint;
                select 
                        BLUID
                into 
                        :KRI_BLUID1 - :KRI_BLUID%left(&KEPT_KRI_BLUID_COUNT)
                from 
                        &blEstimates
                where 
                        CCUID=&&&CCUID&i
                    AND
                        Iteration = &FREQ_ITERATION;
        	quit;
            %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
                %put ERROR: Unable to obtain the list of business lines for row with CCUID=&&&CCUID&i and Iteration = &FREQ_ITERATION;
                %let abort=1;
                %let reason=Unable to obtain the list of business lines for row with CCUID=&&&CCUID&i and Iteration = &FREQ_ITERATION;
                %goto DONE;
            %end;
		%end;  /* &KEPT_KRI_BLUID_COUNT > 0 */

        data _tmpDS_; /* Will have all business lines, even those with estimate=.  Will have >= 1 obs. */
            set &blEstimates;
            where CCUID=&&&CCUID&i
                    AND 
                  Iteration =&FREQ_ITERATION;
        run; 
        %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
            %put ERROR: Unable to create the dataset w/ business lines for this cell.;
            %let abort=1;
            %let reason=Unable to create the dataset w/ business lines for this cell;
            %goto DONE;
        %end;
            
        proc sort data=_tmpDS_ out=_tmpDS_;
            by bluid;
        run;
        %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
            %put ERROR: Unable to sort business line dataset by business line ID.;
            %let abort=1;
            %let reason=Unable to sort business line dataset by business line ID.;
            %goto DONE;
		%end;
    	
		%if &_freqKriAll_exists_ %then %do;

			%if &_only_Firm_KRIs_ %then %do;
				/* Then _freqKriAll has only one observation.  Has variables for:  
				the value of each firm-KRI,
				and RISK_FACTOR_VARIABLE_DATE (date for which we are doing the prediction of the mean/variance) */
				data _tmpDS_;
					if _n_ EQ 1 then set _FreqKriAll_;
					set  _tmpDS_ ;
				run;
			%end;
			%else %do; 
				/* Then _freqKriAll has one observation per BL in this cell.  Has variables for:
				the value of each KRI (firm or BL or both), possibly different for each BL, 
				and RISK_FACTOR_VARIABLE_DATE (date for which we are doing the prediction of the mean/variance) */
	            data _tmpDS_;
	                merge _tmpDS_ (IN=InTmpDS) _FreqKriAll_;
	                if InTmpDS=1;
	                by BLUID; 
	            run;
	            %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	                %put ERROR: Unable to merge _tmpDS_ with business line KRI information dataset.;
	                %let abort=1;
	                %let reason=Unable to merge _tmpDS_ with business line KRI information dataset.;
	                %goto DONE;
				%end;
			%end;
		%end;
		%else %do;  /* we know COUNT_KRI > 0, yet &_freqKriAll_exists == 0; thus, an error */
			%put ERROR: CCUID=&&&CCUID&i: COUNT_KRI>0, yet no KRI information was provided to orva_freq_kri_effect.sas.;
            %let abort=1;
            %let reason=CCUID=&&&CCUID&i: COUNT_KRI>0, yet no KRI information was provided to orva_freq_kri_effect.sas.;
            %goto DONE;
		%end;

        /*- Create the Frequency KRI Model  -*/
        data _null_;
            format TheModel $2000.;
            set &&&DSNAME&i;
				/* Fix (2/2) for wrong way of keeping fits --BEGIN-- */
            where CCUID=&&&CCUID&i AND status=0; /* Only after the AND */
				/* Fix (2/2) for wrong way of keeping fits --END-- */
            TheModel="0";  /* Do not modify or delete this line!!! */
			TheModel = trim(TheModel) || " + " || Intercept;  /* Do not modify or delete this line!!! */
			%if &COUNT_KRI GT 0 %then %do;
                array MyKRIs[&COUNT_KRI] &FREQ_KRI_LIST; /* &FREQ_KRI_LIST will be empty if no KRIs were selected */
				do ki=1 to &COUNT_KRI; /* This loop won't go if no KRIs were selected */
                    TheModel = trim(TheModel) || " + " || MyKRIs(ki) || " * " || vname(MyKRIs(ki));   
                end;
			%end;
            call symputX("TheFreqKriModel", TheModel); 
            call symputX("TheDistribution",Distribution);
            call symputX("TheAlpha",_Alpha);
        run;  
        %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
            %put ERROR: Unable to construct the KRI model string.;
            %let abort=1;
            %let reason=Unable to construct the KRI model string.;
            %goto DONE;
		%end;

		data _TheResults_;
            set _tmpDS_ end=last;
            retain mTot 0;
			/* &TheFreqKriModel will be 0 if no KRIs were selected, which is fine */
			if missing(Estimate) then Mean = exp(&TheFreqKriModel);
			else Mean = exp(&TheFreqKriModel+Estimate);
			if Mean ne . then MTot=Mtot+Mean;
			if last then do;
				call symput("TotalMean", MTot);
			end;	
        run;
        %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
            %put ERROR: Unable to evaluate the model effects.;
            %let abort=1;
            %let reason=Unable to evaluate the model effects.;
            %goto DONE;
        %end;  

        %if &i eq 1 %then %do;
            data _foo_; 
                CCUID=&CCUID1;
				length Distribution $12;
                Distribution = upcase("&TheDistribution");
				/* check whether poisson or negbin */
				if upcase(Distribution)=:'P' then do; /* =: operator does character comparison */
                    FParam_1= &TotalMean; /* the mean of the Poisson distribution */
				end;
				else do;
					FPARAM_1 = &TotalMean / &TheAlpha; /* =n (from SAS documentation for CDF fn for NegBin)*/ 
					FPARAM_2 = &TheAlpha; /* Note that 1/(1+_Alpha) is the probability of success p. */
				end;
            run;
        %end;
        %else %do;
            data _TMP_; 
                CCUID=&&&CCUID&i;
				length Distribution $12;
                Distribution = upcase("&TheDistribution");
				/* check whether poisson or negbin */
				if upcase(Distribution)=:'P' then do; /* =: operator does character comparison */
                    FParam_1= &TotalMean; /* the mean of the Poisson distribution */
				end;
				else do;
					FPARAM_1 = &TotalMean / &TheAlpha; /* =n (from SAS documentation for CDF fn for NegBin)*/ 
					FPARAM_2 = &TheAlpha; /* Note that 1/(1+_Alpha) is the probability of success p. */
				end;
            run;
            data _foo_;
                set _foo_ _tmp_;
            run;                  
        %end;


%NEXTCCUID:
    %end; /* %do i=1 %to &NROWS (Loop through the rows to calculate the KRI & bl effects) */



    data &outFParmEffect(keep= Distribution Fparam_1 FParam_2 CCUID);
        set _foo_;
    run;

    %DONE:

	data &outstatus;
		format status 8. reason $100.;
		status=0;
		reason="";
	run;
	%if &abort %then %do;
        %put ERROR: orva_freq_kri_effect macro ended abnormally.;

        data &outstatus;
            format status 8. reason $100.;
            status=1;
            reason="&reason";
    	run;
    %end;  

    
    proc datasets library=work memtype=data nolist;
        delete _foo_ temp _freqKriAll_ _tmp_ _tmpDS_ _tmpDS0_ _TheResults_ __tmp__ ;
    run;
    quit;    

%mend;
