/*****************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    orva_get_parm_kris                                             
 * PURPOSE: Get the unique list of KRI variables from a parameter dataset by CCUID
 *          and merge with BLUID_KRIs.  Used by the VaRCriteria.
 *        
 * SUPPORT: stchri
 *
 * PARAMETERS:
 *
 *   parmData = the parameter dataset (from fitting).
 *   outData  = output dataset with with NAME column containing kri names.
 *   type     = SEVERITY or FREQUENCY.
 *   ccData   = Cross-Classification dataset for project.
 *   lossData = loss dataset for project.
 *
 * NOTE: see sample code at bottom of file
 *
 *****************************************************************************************/
%macro orva_get_parm_kris(parmData=, outData=, type=SEVERITY, ccData=,
                          lossData=);
   
   /*** create empty output dataset up front ***/ 
   proc sql;
     create table &outData(
          NAME       varchar(32),
          CCUID      num,
          BLUID_KRI  num ); 
   quit;
   
   /*** check parmData ***/
   %let NOBS= 0;
   %let dsid=%sysfunc(open(&parmData,I));
   %if &dsid ^= 0 %then %do;
      %let NOBS=%sysfunc(attrn(&dsid,nlobs));
      %let rc=%sysfunc(close(&dsid));
   %end;
   %if &NOBS eq 0 %then %do;
      %return;
   %end;
	
   /*** transpose parm dataset and keep only KRI variables and CCUID ***/
	proc transpose data= &parmData out=_trans; run;
	proc sort data=_trans; by _NAME_; run;
	data _trans; set _trans;
	  length NAME $ 32;
      array COL[&NOBS] COL:;
	  array CC[&NOBS];
	  retain CC: ;
	  if upcase(substr(_NAME_,1,12)) eq 'CCUID' then do;
	     do i=1 to &NOBS;
		     CC[i] = COL[i];
		 end;
      end;
	  %if &type eq SEVERITY %then %do;
	    if upcase(substr(_NAME_,1,12)) eq 'PARAM_1_KRI_' then do;
	       NAME = trim(substr(_NAME_,9));
		   do i=1 to &NOBS;
		      if COL[i] ne . then do;
		         CCUID = CC[i];
	             output;
		      end;
		   end;
	    end;
	  %end;
	  %else %do;
	    if upcase(substr(_NAME_,1,4)) eq 'KRI_' then do;
		   NAME = _NAME_;
		   do i=1 to &NOBS;
		      if COL[i] ne . then do;
		         CCUID = CC[i];
	             output;
		      end;
		   end;
	    end;
	  %end;
	  keep NAME CCUID;
    run;

	/*** subset loss dataset to get one entry per businessline ***/
	data _losses; set &lossData;
      where AMOUNT ne . and CCVUID_0 ne 0 and CCVUID_1 ne 0;
      keep CCUID CCVUID_: INTERNAL_BL_UID FINANCIALDATE;
    run;
    options sortdup=logical;
    proc sort data=_losses nodups; 
       by CCUID INTERNAL_BL_UID FINANCIALDATE;
    run;
    data _losses; set _losses;
	   by CCUID INTERNAL_BL_UID FINANCIALDATE;
	   if LAST.INTERNAL_BL_UID then output;
    run;

	/*** merge subsetted loss dataset with list of  
	     significant kris by cell (CCUID) ***/
    proc sql; 
      create table _merge as 
        select * from
        _losses l left join _trans t on l.CCUID eq t.CCUID;
    quit;

	/*** drop losses that do not match our significant KRI BLs ***/
    data _merge; set _merge (rename=(INTERNAL_BL_UID=BLUID_KRI)); 
       where NAME ne ""; 
    run;
    data &outData; set _merge(keep=NAME CCUID CCVUID_: BLUID_KRI); run;

    /*** cleanup temporary datasets 
	proc datasets library=work memtype=data nolist nodetails;
	   delete _trans _losses _merge;
	quit;***/
	
%mend orva_get_parm_kris;


/*******************************************************************
%orva_get_parm_kris(parmData=VARPRJ.selsparmi,
                    outData=tmps_pkri,
                    type=SEVERITY,
                    ccData=VARPRJ_M.cc,
                    lossData=VARPRJ_M.losskri);
%orva_get_parm_kris(parmData=VARPRJ.selfparmi,
                    outData=tmpf_pkri,
                    type=FREQUENCY,
                    ccData=VARPRJ_M.cc,
                    lossData=VARPRJ_M.losskri);
********************************************************************/