/*****************************************************************************************
 * $Id: orva_getmomts_scenario_bucketed.sas,v 1.1.2.26.4.3 2006/10/20 15:30:48 stchri Exp $
 * Copyright (C), 2005
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    ORVA_GETMOMTS_SCENARIO_BUCKETED.sas                                             
 * PURPOSE: Get frequency and severity moments for all cells and aggregations for
 *          bucketed scenarios.
 * SUPPORT: sasdje
 *        
 * PARAMETERS:
 *
 *   scenData      =  the name of the scenario dataset with the following variables:
 *                    &CCKEY  BUCKET_UID. The rest formatted as the load tables.
 *   bucketData    =  Description of all the buckets, containing:
 *                    BUCKET_UID. 
 *   ccKey         =  Cross classification key - default is CCUID.
 *   nbuckets      =  Maximum number of buckets
 *   interval      =  DATE interval (YEAR, MONTH ...) Default is YEAR
 *   outSevMomts   =  MEAN, VARIANCE, STD, SKEWNESS, KURTOSIS, COUNT, MINLOSS, MAXLOSS, CCV_n
 *   outFreqMomts  =  MEAN, VARIANCE, CCV_n
 *   outCombined   =  Combined scenario for each cell - needed for simulation and graphing
 *   outScenarios  =  unCombined scenario for each cell - needed for simulation and graphing
 *   outStatus     =  Success or failure - default is status
 *
 *****************************************************************************************/
 %macro ORVA_GetMomts_scenario_bucketed(
                    ScenData     =,
                    BucketData   =,                        
                    ccKey        =CCUID,
                    nbuckets     =10,
                    interval     =YEAR,
                    outSevMomts  =osstat,              
                    outFreqMomts =ofstat,
                    outCombined  =outCombined, 
                    outScenarios =outScenarios,
                    outStatus    =status
                    );

        
        /*- create the status data set up front -*/
        %let status=0;
        %let reason=Successful;
        %let abort=0;
        data &outStatus;
                format status 8. reason $100.;
                status=1;
                reason="Unknown";
        run;
 
        /*- error checking -*/
        %if &ScenData=%str() %then %do;
                %put ERROR: The parameter ScenData= must be specified.;
                %let abort=1;
                %let reason=MISSING ScenData= PARAMETER;
                %goto DONE;
        %end;
                %if &BucketData=%str() %then %do;
                %put ERROR: The parameter BucketData= must be specified.;
                %let abort=1;
                %let reason=MISSING BucketData= PARAMETER;
                %goto DONE;
        %end;
                  
        /*- check for empty dataset -*/
        %let dsid=%sysfunc(open(&ScenData,I));
        %let nLosses=%sysfunc(attrn(&dsid,nlobs));
        %let rc=%sysfunc(close(&dsid));
        %if &nLosses=0 %then %do;
                %put ERROR: No observations available in data.;
                %let abort=1;
                %let reason=No observations available in data;
                %goto DONE;
        %end;

                /*- Merge buckets in each CC ----------------*/  
        proc sql noprint;
             create table _buckets_ as select * from &scendata s, &BucketData b
                      where s.bucket_uid = b.bucket_uid;
        quit;

                

                /*- get a list of unique buckets for each CC -*/
        options sortdup=logical;
        proc sort data=_buckets_(keep= bucket_uid &ccKey) nodup out=_uniquebucket_ids_;
            by bucket_uid &ccKey;
        run; 

        /*- Merge with BucketData -*/
        proc sort data=&BucketData out=_sortedBucketData; by bucket_uid; run;
        data _uniquebucket_ids_; merge _sortedBucketData _uniquebucket_ids_;
             by bucket_uid;
             if &ccKey ^= . then output;
        run;
                   
        /*- Now create a unique set of buckets for each cc -*/ 
        data _uniquebuckets_; set _uniquebucket_ids_;
            array bucket[&nbuckets];

            value = bucket0;
            if value ^= . then output; 
            do i = 1 to &nbuckets;
               value = bucket[i];
               if value ^= . then output;
            end;
            keep value &ccKey;
        run;

        proc sort data=_uniquebuckets_(keep=value &ccKey) nodup out=_uniquebucket_ids_ ;
            by &ccKey value;
        run; 

                    /* Find the new max number of buckets ---------*/
        proc summary data=_uniquebucket_ids_;
                by &cckey;
                var value;
                output out=_nbuckets_ n=nbuckets;
        run;

        proc summary data=_nbuckets_;
                var nbuckets;
                output out=_max_nbuckets_ max=max_nbuckets;
        run;
        data _null_;set _max_nbuckets_; call symput("nbuckets", max_nbuckets);
        run;

        proc transpose data=_uniquebucket_ids_ out=_final_buckets_ prefix=final_bucket;
                by &ccKey;
        run;

                /* add bucket counts to each CC */
        %let nbuckets=%trim(&nbuckets);
        %let nb = %eval(&nbuckets+1);
        data _final_buckets_; set _final_buckets_ ;
           array final_bucket[&nb];
            
           do nbuckets = 1 to &nb until(final_bucket[nbuckets] = .);
           end;
           nbuckets = nbuckets -1;
                       /* shift the buckets backwards --------------*/
           final_bucket0 = final_bucket1;
           do i=1 to nbuckets-1;
              final_bucket[i] = final_bucket[i+1];
           end;
           final_bucket[i] = .;
           nbuckets = nbuckets -1;
           output;
           drop i;
        run;

                /* scale and weight the scenarios so they can be added */
        data _variance_; set &scendata;
           Frequency_Variance = Frequency_Mean; /* poisson by default */
           if Frequency_min ^= . then diffmin = Frequency_mean - Frequency_min;
           if Frequency_max ^= . then diffmax = Frequency_max - Frequency_mean;
           diff = max( diffmin, diffmax);
           if diff ^= . then Frequency_Variance = ( diff/3 ) ** 2;
           keep Frequency_Mean Frequency_Variance &ccKey;
        run;

        proc sort data=_variance_; by &ccKey; run;
        proc summary data=_variance_; 
           by &ccKey;
           var Frequency_Mean Frequency_Variance;
           output out=_frequency_sum_ sum= Frequency_Mean_sum Frequency_Variance_sum;
        run;
        proc sort data=_frequency_sum_; by &cckey;run;
        proc sort data=_buckets_;by &cckey;run;

                 /* scale the pdfs to one * frequency  */
        data &outScenarios; merge _buckets_ _frequency_sum_;
           by &cckey;
           array bucketwgt[&nb];
           array bucket[&nb];

           do nb = 1 to &nb until(bucket[nb] = . or bucketwgt[nb] = .);
           end;
           nb = nb -1;
           sum = 0;
           do i=1 to nb; 
              sum = sum + bucketwgt[i];
           end;
                          /* Wgt each scenario according to their contribution */
           sum = sum * Frequency_mean_sum/Frequency_mean;
           do i=1 to nb;
              bucketwgt[i] =  bucketwgt[i]/sum;
           end;
                  
        run;
        
        /* Now expand the buckets to match combined buckets ------*/
        %orvd_negbin;
        data &outCombined; merge &outScenarios _final_buckets_;
           by &cckey;
           array final_bucketwgt[&nbuckets];
           retain final_bucketwgt1-final_bucketwgt&nbuckets;
           retain ratio;
           array bucketwgt[&nbuckets];
           array bucket[&nbuckets];
           array final_bucket[&nbuckets];
           array nevents[&nbuckets];

                   /* Fix the frequency interval to match the fit interval */
           if _N_ = 1 then do;
              ratio = intck("YEAR", 0, 366)/
              intck("&Interval", 0, 366);
           end; 

           if first.&cckey then do;
              do i = 1 to &nbuckets;
                  final_bucketwgt[i] = 0;
              end;
           end;
           
           curr_b = 1;
           prev_bucket = bucket0;
           prev_final_bucket = final_bucket0;
           do i=1 to nbuckets;
              if bucket[curr_b] = . or bucketwgt[curr_b] = . then i=nbuckets;
              else do;
                  /* Compute portion of current bucket in final bucket */
                  wgt = 1- ((prev_final_bucket-prev_bucket) + 
                         (bucket[curr_b]-final_bucket[i])) / ( bucket[curr_b]-prev_bucket);
                  final_bucketwgt[i] = final_bucketwgt[i] + bucketwgt[curr_b] *wgt;
                  prev_final_bucket = final_bucket[i];
                  if bucket[curr_b] <= final_bucket[i] then do;
                     prev_bucket = bucket[curr_b];
                     curr_b +1;
                  end;
              end;
           end;
                   
           if last.&cckey then do;
               do i=1 to nbuckets;
                      nevents[i] = final_bucketwgt[i] * frequency_mean_sum;
               end;

               mean = frequency_mean_sum;
               variance = frequency_variance_sum;
               if Mean >= Variance then do;
                  Mean = ratio * Mean;
                  Variance = ratio * Variance;
                  FParam_1 = Mean;
                  FParam_2 = .;
                  distribution = "POIS";
               end;
               else do;
                  FParam_2 = Variance/Mean - 1;
                  FParam_1 = ratio * Mean / FParam_2;
                  %MomentsNEGBIN;
                  distribution = "NEGBIN";
               end;
               output;
           end;
           length distribution $ 32;

           keep &cckey final_bucketwgt1-final_bucketwgt&nbuckets 
                nevents1-nevents&nbuckets 
                final_bucket0-final_bucket&nbuckets Mean
                Variance Distribution FParam_1 FParam_2 nbuckets;
        run;


        /* At last we can compute the moments -------------------*/
        data &outSevMomts; set &outCombined;
            array final_bucketwgt[&nbuckets];
            array final_bucket[&nbuckets];

            mean = 0;
            prev_bucket = final_bucket0;
            do i=1 to nbuckets;
               mean = mean + 0.5*(final_bucket[i] + prev_bucket) * 
               final_bucketwgt[i];
               prev_bucket = final_bucket[i];
            end;
                   
            variance = 0;
            skewness = 0;
            kurtosis = 0;
            prev_bucket = final_bucket0;
            do i=1 to nbuckets;
               tmp = (final_bucket[i]-mean)**3 - ( prev_bucket - mean ) **3;
               variance = variance + final_bucketwgt[i]*tmp/3 /
                                    (final_bucket[i] - prev_bucket); 
               prev_bucket = final_bucket[i];
            end;
            STD = sqrt(variance);
            prev_bucket = final_bucket0;
            do i=1 to nbuckets;
               tmp = (final_bucket[i]-mean)/std;
               tmp2 =(prev_bucket - mean)/std;
               f = final_bucketwgt[i]/
                                    (final_bucket[i] - prev_bucket);
               skewness = skewness + 1/4 * f * ( tmp **4 - tmp2 **4);
               kurtosis = kurtosis + 1/5 * f * ( tmp **5 - tmp2 **5);
               prev_bucket = final_bucket[i];
               if i eq nbuckets then MAXLOSS = final_bucket[i];
            end;
            skewness = std * skewness;
            kurtosis = std * kurtosis;
                   
            keep Variance Skewness Kurtosis Mean Std &ccKey MAXLOSS;
        run;
        
        data &outFreqMomts; set &outCombined;
                Keep Mean Variance &ccKey;
        run;
                
        data &outScenarios;
           set &outScenarios;
           drop i prev_bucket sum bucketwgt&nb bucket&nb;
           array bucketwgt[&nbuckets];
           array nevents[&nbuckets];
           array final_bucketwgt[&nbuckets];
           do i=1 to nb;
                      final_bucketwgt[i] = bucketwgt[i];
                      nevents[i] = final_bucketwgt[i] * frequency_mean_sum;
           end;
           rename nb=nbuckets; 
        run;
          
            /*- finalize status dataset -*/
%DONE:
    data &outStatus;
                format status 8. reason $100.;
                status=0;
                reason="";
    run;
    
        %if &abort %then %do;
                %put ERROR: orva_getmomts macro ended abnormally.;

                data &outstatus;
                        format status 8. reason $100.;
                        status=1;
                        reason="&reason";
                run;
        %end;  
  
    proc datasets library=work memtype=data nolist;
         delete _buckets_ _variance_ _final_buckets_ _frequency_sum_
                _uniquebucket_ids_ _uniquebuckets_ _tmpData _sortedBucketData;
    run; quit;
                 
%mend;

  
