 /*******************************************************************
  * Copyright (C), 2004
  * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
  *
  * Title: Anderson-Darling, Kolmogorov-Smirnov, 
  *        Cramer-Von Mises goodness of fit tests
  *
  * Support: jariou - Jacques Rioux
  *
  * Authors: Vijay Desai & Bill Szczuka, SAS San Diego
  *          Apr. 2003
  *
  * Parameters:
  *   data             =  Name of dataset to be analyzed
  *   var              =  Name of variable to be analyzed
  *   thresholdVar     =  Name of threshold variable
  *   ccUid            =  Current CCUID (cell key)
  *   dist             =  Distribution to be fit
  *   parmdata         =  Name of parameter data set
  *   gofThreshold     =  Goodness of Fit threshoold
  *   defaultThreshold =  Default threshold
  *
  * Censored GoodnessOfFit values are placed in global macros:
  *   AndersonDarling_c, KolmogorovSmirnov_c, CramerVonMises_c
  *
  * The respective p Value ranges will be placed in:
  *   AndersonDarling_p, KolmogorovSmirnov_p, CramerVonMises_p
  *
  * The corresponding stat value for each p Value range:
  *   AndersonDarling_l, AndersonDarling_r
  *   KolmogorovSmirnov_l, KolmogorovSmirnov_r
  *   CramerVonMises_l, CramerVonMises_r
  *
  * The Interpolated p value:
  *   AndersonDarling_i, KolmogorovSmirnov_i, CramerVonMises_i
  *
  * The Uncensored Goodness of Fit stats:
  *   AndersonDarling_u, KolmogorovSmirnov_u, CramerVonMises_u
  *
  ******************************************************************/
%macro ORVA_GFit(data=, 
				 var=, 
				 thresholdVar=,
				 ccUid=,
				 dist=, 
				 parmdata=, 
				 gofThreshold=,
				 defaultThreshold=0
				 )  ;
				 
    %global AndersonDarling_u KolmogorovSmirnov_u CramerVonMises_u
		    AndersonDarling_c KolmogorovSmirnov_c CramerVonMises_c
		    AndersonDarling_p KolmogorovSmirnov_p CramerVonMises_p
		    AndersonDarling_i KolmogorovSmirnov_i CramerVonMises_i
		    AndersonDarling_l KolmogorovSmirnov_l CramerVonMises_l
		    AndersonDarling_r KolmogorovSmirnov_r CramerVonMises_r;
		
	%let dist=%upcase(&dist);
	%orvd_&dist /* force load of macros */
	
	%let abort=0;

	%let AndersonDarling_u=;
	%let AndersonDarling_c=;
	%let AndersonDarling_p=;
	%let AndersonDarling_i=;
	%let AndersonDarling_l=;
	%let AndersonDarling_r=;
	
	%let KolmogorovSmirnov_u=;
	%let KolmogorovSmirnov_c=;
	%let KolmogorovSmirnov_p=;
	%let KolmogorovSmirnov_i=;
	%let KolmogorovSmirnov_l=;
	%let KolmogorovSmirnov_r=;
	
	%let CramerVonMises_u=;
	%let CramerVonMises_c=;
	%let CramerVonMises_p=;
	%let CramerVonMises_i=;
	%let CramerVonMises_l=;
	%let CramerVonMises_r=;
	

	data _parm_;
		set &parmdata(keep=_numeric_);
	run;

	proc transpose data=_parm_ out=_parm_;
	run;

	data _null_;
		set _parm_;
		call symput(_name_, col1);
	run;

    /*Using gfit_data for the dataset to use to calculate goodness of fit*/
	%if &gofThreshold=%str() %then %do;
		data gfit_data;
			set &data; 
			where CCUID=&ccUid;
			if &THRESHOLDVar=. then &THRESHOLDVar=&defaultThreshold;
		run;
	%end;
	%else %do;
        data gfit_data;
            set &data; 
            where CCUID=&ccUid;
			&THRESHOLDVar=&gofThreshold;
		run;
	%end;
  
	/*- data sanity check -*/
	data _null_;
        set gfit_data nobs=nobs;
		retain NExcess 0;
	
		if &var>&THRESHOLDVar then NExcess=NExcess+1;
		
		if _N_=nobs then do;
			call symput("NExcess", NExcess);   
		end;
	run;  


	%if &NExcess<1 %then %do;
		%put ERROR: There are no observations above the test threshold.;
		%let abort=1;
		%let reason = 'There are no observations above the test threshold.';
		%goto DONE;
	%end;

	%let NumRowsAndersonDarlingTable=112;
	%let NumColsKS_Censored=10;
	%let NumColsCVM_Censored=9;
	%let NumColsAD_Censored=9;

	data _temp_;
			set gfit_data nobs=NGoodnessFit;

			if &THRESHOLDVar=0 then do;
				truncdf=0;
			end;
			else do;
				%Cdf&dist(&THRESHOLDVar,truncdf);
			end;

			%Cdf&dist(&var,F)
			F=(F-truncdf)/(1-truncdf);    
		run;	


   proc sort data=_temp_ out=_temp_(where=(&var>&THRESHOLDVar));
		by F;
	run;


	data _temp_(keep=KS_Stat_Censored CVM_Stat_Censored AD_Stat_Censored  N s);
		set _temp_ nobs=NGoodnessFit;
		retain s x d j DRN WRN1 ARN1 ARN2 0;

		N=NGoodnessFit+1;
		DRN=max(DRN,abs((_N_-.5)/N-f)+.5/N);
		WRN1=WRN1+(f-(2*_N_-1)/(2*N))**2;
       ARN1=ARN1+(2*_N_-1)*log(max(f,1e-28))+(2*NGoodnessFit+1-2*_N_)*log(max(1-f,1e-28));
		ARN2=ARN2+log(max(1-f,1e-28));
		
        /*--- Anderson-Darling ---*/
        s= s+((2*_N_-1)*log(max(f,1e-28)) +(2*NGoodnessFit+1-2*_N_)* log(max(1-f,1e-28)))/NGoodnessFit; 

        /*--- Kolmogorov-Smirnov ---*/
        d=max(abs(f-_N_/NGoodnessFit),d);                  

        /*--- Cramer-Von Mises ---*/ 
        x=x+((f)-(2*_N_-1)/(2*NGoodnessFit))**2;           

		if _N_=NGoodnessFit then do;
			KS_Stat_Censored=round(DRN,0.0001);
			CVM_Stat_Censored=round(WRN1+_N_/(12*N**2)+(N/3)*(f-_N_/N)**3,0.0001);
			if (0<=f<=1) then do;
				AD_Stat_Censored = round(-(1/n)*ARN1-2*ARN2-(1/N)*( (_N_-N)**2*log(max(1-f,1e-28))-_N_**2*log(max(f,1e-28))+N**2*f ),0.0001);
			end;
		end; /*- end of _N_=NGoodnessFit -*/
		else do;
			delete;
		end;
	run;
	
	data _null_;
		set _temp_;
        length 
        AndersonDarling_p   
        AndersonDarling_l   
        AndersonDarling_r
        KolmogorovSmirnov_p 
        KolmogorovSmirnov_l 
        KolmogorovSmirnov_r
        CramerVonMises_p    
        CramerVonMises_l    
        CramerVonMises_r $40;
       
		array KS_Censored_Sig (&NumColsKS_Censored,1) _temporary_ 	(1,.50,.25,.15,.10,.05,.025,.01,.005,0);
		array KS_Censored_2 (&NumColsKS_Censored,1) _temporary_  	(0,.4923,.6465,.7443,.8155,.9268,1.0282,1.1505,1.2361,10);
		array KS_Censored_3 (&NumColsKS_Censored,1) _temporary_  	(0,.5889,.7663,.8784,.9597,1.0868,1.2024,1.3419,1.4394,10);
		array KS_Censored_4 (&NumColsKS_Censored,1) _temporary_ 	(0,.6627,.8544,.9746,1.0616,1.1975,1.3209,1.4696,1.5735,10);
		array KS_Censored_5 (&NumColsKS_Censored,1) _temporary_ 	(0,.7204,.9196,1.0438,1.1334,1.2731,1.3997,1.5520,1.6583,10);
		array KS_Censored_6 (&NumColsKS_Censored,1) _temporary_ 	(0,.7649,.9666,1.0914,1.1813,1.3211,1.4476,1.5996,1.7056,10);
		array KS_Censored_7 (&NumColsKS_Censored,1) _temporary_ 	(0,.7975,.9976,1.1208,1.2094,1.3471,1.4717,1.6214,1.7258,10);
		array KS_Censored_8 (&NumColsKS_Censored,1) _temporary_ 	(0,.8183,1.0142,1.1348,1.2216,1.3568,1.4794,1.6272,1.7306,10);
		array KS_Censored_9 (&NumColsKS_Censored,1) _temporary_ 	(0,.8270,1.0190,1.1379,1.2238,1.3581,1.4802,1.6276,1.7308,10);
		array KS_Censored_10 (&NumColsKS_Censored,1) _temporary_ 	(0,.8276,1.0192,1.1379,1.2238,1.3581,1.4802,1.6276,1.7308,10);

		array CVM_Censored_Sig (&NumColsCVM_Censored,1) _temporary_ (1,.50,.25,.15,.10,.05,.025,.01,0);
		array CVM_Censored_10 (&NumColsCVM_Censored,1) _temporary_ 	(0,.117,.212,.288,.349,.456,.564,.709,5);
		array CVM_Censored_20 (&NumColsCVM_Censored,1) _temporary_ 	(0,.116,.212,.288,.350,.459,.572,.724,5);
		array CVM_Censored_40 (&NumColsCVM_Censored,1) _temporary_ 	(0,.115,.211,.288,.350,.461,.576,.731,5);
		array CVM_Censored_100 (&NumColsCVM_Censored,1) _temporary_ (0,.115,.211,.288,.351,.462,.578,.736,5);
		array CVM_Censored_999 (&NumColsCVM_Censored,1) _temporary_ (0,.119,.209,.284,.347,.461,.581,.743,5);			

		array AD_Censored_Sig (&NumColsAD_Censored,1) _temporary_ 	(1,.50,.25,.15,.10,.05,.025,.01,0);
		array AD_Censored_10 (&NumColsAD_Censored,1) _temporary_ 	(0,.775,1.248,1.610,1.933,2.492,3.070,3.857,20);


			if KS_Stat_Censored ne . then do;
				r_over_N=round(_N_/N,.1);

				%if r_over_N=1 %then %do;
					%orva_binarysearch(ArrayXY=KS_Censored_10, size=&NumColsKS_Censored, value=KS_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
				%end; 
				%else %if r_over_N=.9 %then %do;
					%orva_binarysearch(ArrayXY=KS_Censored_9, size=&NumColsKS_Censored, value=KS_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
				%end; 
				%else %if r_over_N=.8 %then %do;
					%orva_binarysearch(ArrayXY=KS_Censored_8, size=&NumColsKS_Censored, value=KS_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
				%end; 
				%else %if r_over_N=.7 %then %do;
					%orva_binarysearch(ArrayXY=KS_Censored_7, size=&NumColsKS_Censored, value=KS_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
				%end; 
				%else %if r_over_N=.6 %then %do;
					%orva_binarysearch(ArrayXY=KS_Censored_6, size=&NumColsKS_Censored, value=KS_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
				%end; 
				%else %do;
					%orva_binarysearch(ArrayXY=KS_Censored_5, size=&NumColsKS_Censored, value=KS_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
				%end;

				L_I=min(&NumColsKS_Censored,L_I);
				R_I=min(&NumColsKS_Censored,R_I);

				if L_I=R_I and (0<L_I<&NumColsKS_Censored) then do;
					KS_Stat_Censored_i=KS_Censored_Sig(L_I,1);
					KolmogorovSmirnov_p="p Value=" || trim(left(KS_Stat_Censored_i));
				end; 
				else do;
					if L_V=R_V then KS_Stat_Censored_i=KS_Censored_Sig(L_I,1);
						else KS_Stat_Censored_i=KS_Censored_Sig(L_I,1)+(KS_Censored_Sig(R_I,1)-KS_Censored_Sig(L_I,1))*((KS_Stat_Censored-L_V)/(R_V-L_V));
							if R_I=&NumColsKS_Censored then do;
								KolmogorovSmirnov_p="p Value<" || trim(left(KS_Censored_Sig(R_I-1,1)));
								KolmogorovSmirnov_l="  (p Value=" || trim(left(KS_Censored_Sig(R_I-1,1))) || " when KS=" || trim(left(L_V)) || ")";
							end; 
							else if L_I=1 then do;
								KolmogorovSmirnov_p="p Value>" || trim(left(KS_Censored_Sig(L_I+1,1)));
								KolmogorovSmirnov_l="  (p Value=" || trim(left(KS_Censored_Sig(L_I+1,1))) || " when KS=" || trim(left(R_V)) || ")";
							end; 
							else do;
								KolmogorovSmirnov_p=trim(left(KS_Censored_Sig(L_I,1))) || ">p Value>" || trim(left(KS_Censored_Sig(R_I,1)));
								KolmogorovSmirnov_l="  (p Value=" || trim(left(KS_Censored_Sig(L_I,1))) || " when KS=" || trim(left(L_V)) || ")";
								KolmogorovSmirnov_r="  (p Value=" || trim(left(KS_Censored_Sig(R_I,1))) || " when KS=" || trim(left(R_V)) || ")";
							end;
						end;
				end; /*- end of KS_Stat_Censored ne . -*/


				if CVM_Stat_Censored ne . then do;
				%if N>300 %then %do;
						%orva_binarysearch(ArrayXY=CVM_Censored_999, size=&NumColsCVM_Censored, value=CVM_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
				%end; 
				%else %if N>70 %then %do;
					%orva_binarysearch(ArrayXY=CVM_Censored_100, size=&NumColsCVM_Censored, value=CVM_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
				%end; 
				%else %if N>30 %then %do;
					%orva_binarysearch(ArrayXY=CVM_Censored_40, size=&NumColsCVM_Censored, value=CVM_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
				%end; 
				%else %if N>15 %then %do;
					%orva_binarysearch(ArrayXY=CVM_Censored_20, size=&NumColsCVM_Censored, value=CVM_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
				%end; 
				%else %do;
				%orva_binarysearch(ArrayXY=CVM_Censored_10, size=&NumColsCVM_Censored, value=CVM_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
				%end;

		L_I=min(&NumColsCVM_Censored,L_I);
		R_I=min(&NumColsCVM_Censored,R_I);

	  if L_I=R_I and (0<L_I<&NumColsCVM_Censored) then do;
		 CVM_Stat_Censored_i=CVM_Censored_Sig(L_I,1);
		 CramerVonMises_p="p Value=" || trim(left(CVM_Stat_Censored_i));
	  end; 
	  else do;
		 if L_V=R_V then CVM_Stat_Censored_i=CVM_Censored_Sig(L_I,1);
		 else CVM_Stat_Censored_i=CVM_Censored_Sig(L_I,1)+(CVM_Censored_Sig(R_I,1)-CVM_Censored_Sig(L_I,1))*((CVM_Stat_Censored-L_V)/(R_V-L_V));

		 if R_I=&NumColsCVM_Censored then do;
			CramerVonMises_p="p Value<" || trim(left(CVM_Censored_Sig(R_I-1,1)));
			CramerVonMises_l="  (p Value=" || trim(left(CVM_Censored_Sig(R_I-1,1))) || " when CVM=" || trim(left(L_V)) || ")";
		 end; 
	 else if L_I=1 then do;
			CramerVonMises_p="p Value>" || trim(left(CVM_Censored_Sig(L_I+1,1)));
			CramerVonMises_l="  (p Value=" || trim(left(KS_Censored_Sig(L_I+1,1))) || " when CVM=" || trim(left(R_V)) || ")";
		 end; 
	 else do;
			CramerVonMises_p=trim(left(CVM_Censored_Sig(L_I,1))) || ">p Value>" || trim(left(CVM_Censored_Sig(R_I,1)));
			CramerVonMises_l="  (p Value=" || trim(left(CVM_Censored_Sig(L_I,1))) || " when CVM=" || trim(left(L_V)) || ")";
			CramerVonMises_r="  (p Value=" || trim(left(CVM_Censored_Sig(R_I,1))) || " when CVM=" || trim(left(R_V)) || ")";
		 end;
	  end;
	end; 
	/*- end of  CVM_Stat_Censored ne . -*/


	if AD_Stat_Censored ne . then do;
		%orva_binarysearch(ArrayXY=AD_Censored_10, size=&NumColsAD_Censored, value=AD_Stat_Censored, left=L_I, right=R_I, left_v=L_V, right_v=R_V)
		L_I=min(&NumColsAD_Censored,L_I);
		R_I=min(&NumColsAD_Censored,R_I);

	  if L_I=R_I and (0<L_I<&NumColsAD_Censored) then do;
		 AD_Stat_Censored_i=AD_Censored_Sig(L_I,1);
		 AndersonDarling_p="p Value=" || trim(left(AD_Stat_Censored_i));
	  end; 
	  else do;
					if L_V=R_V then do;
						AD_Stat_Censored_i=AD_Censored_Sig(L_I,1);
					end;
					else do;
						AD_Stat_Censored_i=AD_Censored_Sig(L_I,1)+(AD_Censored_Sig(R_I,1)-AD_Censored_Sig(L_I,1))*((AD_Stat_Censored-L_V)/(R_V-L_V));
					end;

		 if R_I=&NumColsAD_Censored then do;
			AndersonDarling_p="p Value<" || trim(left(AD_Censored_Sig(R_I-1,1)));
			AndersonDarling_l="  (p Value=" || trim(left(AD_Censored_Sig(R_I-1,1))) || " when AD=" || trim(left(L_V)) || ")";
		 end; 
	 else if L_I=1 then do;
			AndersonDarling_p="p Value>" || trim(left(AD_Censored_Sig(L_I+1,1)));
			AndersonDarling_l="  (p Value=" || trim(left(AD_Censored_Sig(L_I+1,1))) || " when AD=" || trim(left(R_V)) || ")";
		 end; 
	 else do;
			AndersonDarling_p=trim(left(AD_Censored_Sig(L_I,1))) || ">p Value>" || trim(left(AD_Censored_Sig(R_I,1)));
			AndersonDarling_l="  (p Value=" || trim(left(AD_Censored_Sig(L_I,1))) || " when AD=" || trim(left(L_V)) || ")";
			AndersonDarling_r="  (p Value=" || trim(left(AD_Censored_Sig(R_I,1))) || " when AD=" || trim(left(R_V)) || ")";
		 end;
	  end;
   end; /*- end of AD_Stat_Censored ne . -*/

			/* AndersonDarlingStat = -NGoodnessFit-s; */
			AndersonDarlingStat = -(N-1)-s;
			
	
   call symput('AndersonDarling_u',trim(left(AndersonDarlingStat)));
   call symput('AndersonDarling_c',trim(left(AD_Stat_Censored)));
   call symput('AndersonDarling_i',trim(left(AD_Stat_Censored_i)));
   call symput('AndersonDarling_p',AndersonDarling_p);
   call symput('AndersonDarling_l',AndersonDarling_l);
   call symput('AndersonDarling_r',AndersonDarling_r);

   call symput('KolmogorovSmirnov_u',trim(left(d)));
   call symput('KolmogorovSmirnov_c',trim(left(KS_Stat_Censored)));
   call symput('KolmogorovSmirnov_i',trim(left(KS_Stat_Censored_i)));
   call symput('KolmogorovSmirnov_p',KolmogorovSmirnov_p);
   call symput('KolmogorovSmirnov_l',KolmogorovSmirnov_l);
   call symput('KolmogorovSmirnov_r',KolmogorovSmirnov_r);

   call symput('CramerVonMises_u',trim(left(x+1/(12*NGoodnessFit))));
   call symput('CramerVonMises_c',trim(left(CVM_Stat_Censored)));
   call symput('CramerVonMises_i',trim(left(CVM_Stat_Censored_i)));
   call symput('CramerVonMises_p',CramerVonMises_p);
   call symput('CramerVonMises_l',CramerVonMises_l);
   call symput('CramerVonMises_r',CramerVonMises_r);
    run;
		
    /*- add these statistics to parameter estimate data set -*/
    data &parmdata;
    set &parmdata;
    /*      AD_V=&AndersonDarling_c;*/
		AD_V=&AndersonDarling_u;
	    AD_P=&AndersonDarling_i;
    	CVM_V=&CramerVonMises_c;
	    CVM_P=&CramerVonMises_i;  
    	KS_V=&KolmogorovSmirnov_c;
	    KS_P=&KolmogorovSmirnov_i;
    run;

    %DONE:
    %if &abort %then %do;
	%put ERROR: The ORVA_GFit macro ended abnormally.;
    %end;

    proc datasets library=work memtype=data nolist;
    delete _temp_ _parm_ gfit_data;
    run;
    quit;

%mend ORVA_GFit;
