/******************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    ORVA_Insurance_Coverage.sas                                             
 * PURPOSE: Creates two datasets giving the probability of primary 
 *          or seconday coverage as well as a coverage proportion
 *          indicating how the recovery dollars are distributed among
 *          business lines
 *
 * SUPPORT: jariou - Jacques Rioux
 *
 * Parameters:
 *
 * LOSSDATA      = Specifies the name of the input dataset to be analyzed.
 *                 Must contain the following columns:
 *                 &CCKEY         Cross classification key  
 *				   INSURANCE_1    Primary insurance policy UID
 *                 INSURANCE_2	  Secondary insurance policy UID
 *
 * CCKEY          = Specifies the cross classification "cell"
 *
 * INSURANCEDATA  = Specifies the name of the insurance policies dataset
 *						  Must contain the following columns:
 *					        INSURANCE_UID
 *
 *						  May also contain:
 *						     DEDUCTIBLE 
 *						     LIMIT
 *						     AGG_LIMIT
 * VERSION = Version of Data for losses that should be used.
 * 
 * COVERAGEDATA   = Prefix for the name of the two datasets created:
 *                  CoverageData_1:   Primary coverage information 
 *                  CoverageData_2:   Secondary coverage information 
 *
 * INSSTATUS      = Name of status dataset indicating success of failure of 
 *					     the macro with additional information on the reason of the failure
 ******************************************************************************/
%macro orva_Insurance_Coverage(LossData=,      
							   ccKey=,   
							   InsuranceData=,
							   CoverageData=,
	    					   insStatus=insStatus,
	    					   version=   
						       );

	/*** set up the status macro variables	***/
	%global status reason;
	%let status=0;
	%let reason=Successful;	
	%let abort=0;

	/*** create the status data set upfront **/
	data &insStatus;
		format status 8. reason $100.;
		status=1;
		reason="Unknown";
		count=0;
	run;


       /* Check for the no internal data case ---------*/
       %orvl_ds_exists(&LossData, "&LossData", false);
          %if &ds_nobs eq 0 %then %do;
            %orvl_ds_exists(VARPRJ.CC, "VARPRJ.CC", false);
            %if &ds_nobs eq 0 %then %do; 
                %let cc_ds = VARPRJ_M.CC;
            %end;
            %else %do;
                %let cc_ds = VARPRJ.CC;
            %end;

            data &CoverageData._1; set &cc_ds(keep=ccuid);
               INSURANCE_1 = 0;
               PIns1 = 1;
               PInsCurrent1 = 1;
            run;
            data &CoverageData._2; set &cc_ds(keep=ccuid);
               INSURANCE_1 = 0;
               INSURANCE_2 = 0;
               PIns2 = 1;
               PInsCurrent2 = 1;
            run;
         %goto DONE;
       %end;

	/*- Replacing missing data by zeroes in the Loss dataset -*/
	data _lossX (compress=binary);
		set &LossData;
        keep &cckey Insurance_1 Insurance_2 Insurance_Recovery_1 Insurance_Recovery_2;
		if INSURANCE_1=. then do;
			INSURANCE_1=0;
		end;
		if INSURANCE_2=. then do;
			INSURANCE_2=0;
		end;
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to replace missing data by zeroes in the &LossData dataset.;
	    %let abort=1;
	    %let reason=Unable to replace missing data by zeroes in the &LossData dataset;
	    %goto DONE;
	%end;

	/*- Extend the insurance dataset by adding a row with a zero Insurance_UID -*/
	data _insuranceX;
		set &InsuranceData nobs=count; where version_UID = &version;
		keep INSURANCE_UID DEDUCTIBLE LIMIT AGG_LIMIT;
		output;		
		if _N_=count then do;
            INSURANCE_UID                 = 0;
            DEDUCTIBLE                    = .;
            LIMIT                         = .;
            AGG_LIMIT                     = .;
 			output;
		end;
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to add a row with a zero Insurance_UID.;
	    %let abort=1;
	    %let reason=Unable to add a row with a zero Insurance_UID;
	    %goto DONE;
	%end;

	/*- Reproduce the extended insurance dataset with different names for secondary coverage -*/
	data _insurance2(rename=(INSURANCE_UID=INSURANCE_UID2  
	                         Deductible=Deductible2
	                         Limit=Limit2 
	                         AGG_Limit=AGG_Limit2)) ;
		set _insuranceX;	
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to create insurance data for secondary coverage.;
	    %let abort=1;
	    %let reason=Unable to create insurance data for secondary coverage.;
	    %goto DONE;
	%end;

	/* Denormalize data by combining the losses with the insurance information */
	proc sql;
		create table _lossTmp 
		as
		select * from 
				_lossX a left join _insuranceX b 
		on	
				a.Insurance_1 = b.Insurance_UID
		left join _insurance2 c 
                on 
				a.Insurance_2 = c.Insurance_UID2
		;
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to denormalize insurance data.;
	    %let abort=1;
	    %let reason=Unable to denormalize insurance data;
	    %goto DONE;
	%end;

	/* Sort the denormalize data by ccuid, Insurance_1 and Insurance_2 */
	proc sort data=_lossTmp;
        by  &cckey 
            Insurance_1
            insurance_2
            ;
	run;
	data _lossTmp (compress=binary); set _lossTmp; run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to sort the denormalize data. ; 
	    %let abort=1;
	    %let reason=Unable to sort the denormalize data.;
	    %goto DONE;
	%end;


	/* Find the number of covered losses for each type of coverage*/
	proc summary data=_lossTmp;
        by  &cckey 
            Insurance_1 
            Insurance_2
            ;
        var Insurance_Recovery_1 Insurance_Recovery_2;
	    output N=Count_1 Count_2 
	           out=_coverage;
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to find the number of covered losses for each type of coverage.; 
	    %let abort=1;
	    %let reason=Unable to find the number of covered losses for each type of coverage;
	    %goto DONE;
	%end;


	/* Find totals by CCUID */
	proc summary data=_coverage;
		by &cckey ;
		var _FREQ_;
        output 
            sum=&CCKey.Sum 
            out=_coverage_CCUID(keep =&cckey &cckey.sum);
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to find totals by &CCKEY.; 
	    %let abort=1;
	    %let reason=Unable to find totals by &CCKEY;
	    %goto DONE;
	%end;

	/* Find totals by CCUID and Insurance_1*/
	proc summary data=_coverage;
		by &cckey Insurance_1 ;
		var _FREQ_;
        output 
                sum=InsSum 
                out=_coverage_INS(keep=&cckey insurance_1 InsSum);
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to find totals by &CCKEY and Insurance_1.; 
	    %let abort=1;
	    %let reason=Unable to find totals by &CCKEY and Insurance_1;
	    %goto DONE;
	%end;


    /* Compute probabilities of primary coverage */
    data &CoverageData._1 (keep=&cckey Insurance_1 PIns1 PInsCurrent1);
		merge 	_coverage_INS _coverage_CCUID;
		by &cckey ;

        if &CCKEY.Sum NE . then do;
            PIns1=rounde(InsSum/&CCKEY.Sum, 0.00001);
        end;
        else do;
            PIns1=0;
		end;
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to compute probabilities of primary coverage.; 
	    %let abort=1;
        %let reason=Unable to compute probabilities of primary coverage;
	    %goto DONE;
	%end;

    data &CoverageData._1 (keep=&cckey Insurance_1 PIns1 PInsCurrent1 );
        set &CoverageData._1;
        by &cckey ;
        retain PIns1 0 TotalPIns1 0 MaxPIns1 0 MaxPIns1_Index 0 KeyIndex 0;

        if First.&cckey then do;
            TotalPIns1 = 0;
            MaxPIns1   = 0;
            KeyIndex= KeyIndex + 1;
        end; 

        TotalPIns1 = TotalPIns1 + PIns1;

        if PIns1 > MaxPIns1 then do;
            MaxPIns1       = PIns1;
            MaxPIns1_Index = _N_;
        end;

        if Last.&cckey then do;
            call symput('TotalPIns1' || left(keyIndex),TotalPIns1);
            call symput('MaxPIns1_Index' || left(keyIndex),MaxPIns1_Index);
        end; 

        PInsCurrent1 = PIns1 ;
    run;


    data &CoverageData._1 (keep=&cckey Insurance_1 PIns1 PInsCurrent1);
        set &CoverageData._1; 
        retain KeyIndex 0 TheKeyIndex "00000";
        by &cckey;

        if First.&cckey then do;
            KeyIndex = KeyIndex + 1;
            TheKeyIndex = KeyIndex;
        end;

        if symget('TotalPIns1' || left(KeyIndex)) > 1.0 then do; 
            if _N_ = symget('MaxPIns1_Index' || left(KeyIndex)) then do;
                tmp=symget('TotalPIns1' || left(KeyIndex))-PIns1;
                PIns1 = rounde(PIns1 - symget('TotalPIns1' || left(KeyIndex)) + 1.0, 0.00001);

                if PIns1 + tmp > 1.0 then do;
                    PIns1 = PIns1-0.00001;
                end;

                PInscurrent1=PIns1;
            end;
        end;
    run;


	/* Merge primary data with CCUID information*/
	data Final_tmp;
		merge 	_coverage_INS _coverage_CCUID;
		by &cckey ;
	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
	    %put ERROR: Unable to merge primary data with &CCKEY information.; 
	    %let abort=1;
	    %let reason=Unable to merge primary data with &CCKEY information;
	    %goto DONE;
	%end;


    /* Compute probabilities of secondary coverage */
    data &CoverageData._2 
                        (keep = &cckey 
                                Insurance_1 
                                Insurance_2 
                                PIns2 
                                PInsCurrent2
                        );
		merge 	Final_tmp _coverage;
		by &cckey insurance_1 ;
        PIns2=rounde(_FREQ_/InsSum, 0.00001);

	run;
	%if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to compute probabilities of secondary coverage.; 
	    %let abort=1;
        %let reason=Unable to compute probabilities of secondary coverage;
	    %goto DONE;
	%end;

    /* Make sure probabilities do not add up to more than 1  */
    data &CoverageData._2 
                        (keep = &cckey 
                                Insurance_1 
                                Insurance_2 
                                PIns2 
                                PInsCurrent2
                        );

        set &CoverageData._2;
        by &cckey Insurance_1;

        retain PIns2 0 TotalPIns2 0 MaxPIns2 0 MaxPIns2_Index 0 KeyIndex 0;

        if First.Insurance_1 then do;
            TotalPIns2 = 0;
            MaxPIns2   = 0;
            KeyIndex   = KeyIndex + 1;
        end; 

        TotalPIns2 = TotalPIns2 + PIns2;

        if PIns2 > MaxPIns2 then do;
            MaxPIns2       = PIns2;
            MaxPIns2_Index = _N_;
        end;

        if Last.Insurance_1 then do;
            call symput('TotalPIns2' || left(keyIndex),TotalPIns2);
            call symput('MaxPIns2_Index' || left(keyIndex),MaxPIns2_Index);
        end; 

        PInsCurrent2 = PIns2 ;
    run;


    data &CoverageData._2(keep = &cckey 
                                Insurance_1 
                                Insurance_2 
                                PIns2 
                                PInsCurrent2
                                );

        set &CoverageData._2; 
        retain KeyIndex 0 TheKeyIndex "00000";
        by &cckey Insurance_1;

        if First.Insurance_1 then do;
            KeyIndex = KeyIndex + 1;
            TheKeyIndex=KeyIndex;
        end;

        if symget('TotalPIns2' || left(KeyIndex)) > 1.0 then do; 
            if _N_ = symget('MaxPIns2_Index' || left(KeyIndex)) then do;
                tmp=symget('TotalPIns2' || left(KeyIndex))-PIns2;
                PIns1 = rounde(PIns2 - symget('TotalPIns2' || left(KeyIndex)) + 1.0, 0.00001);

                if PIns2 + tmp > 1.0 then do;
                    PIns2 = PIns2 - 0.00001;
                end;

                PInscurrent2=PIns2;
            end;
        end;
    run;


%DONE:

	%if &abort %then %do;
	    %put ERROR: orva_insurance_coverage macro ended abnormally.;
	    %let status=1;
	%end;

	/*- delete unwanted data -*/
	proc datasets lib=work memtype=data nolist;
		delete  _Summ_:
				_coverage 
                _coverage_: 
				Final_tmp 
				_loss:
				_insurance:
				;
	run;
	quit;

	data &insStatus;
		format status 8. reason $100.;
		status=&status;
		reason="&reason";
	run;
%mend orva_Insurance_Coverage;

