/******************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    ORVA_Insurance_Proportion.sas                                             
 *
 * PURPOSE: Calculates the coverage proportion of each insurance contract 
 *          for each  CCUID. This is used to allocate the aggregate limit 
 *          of each contract per cell
 *
 * SUPPORT: jariou - Jacques Rioux
 *
 * Parameters:
 *
 * LOSSDATA      = Specifies the name of the input data set to be analyzed.
 *                 Must contain the following columns:
 *                 &CCKEY          Cross classification key  
 *                 INSURANCE_1     Primary insurance policy UID
 *                 INSURANCE_2     Secondary insurance policy UID
 *
 * CCKEY          = Specifies the cross classification "cell"
 *
 * INSURANCEDATA  = Specifies the name of the insurance policies dataset
 *                      Must contain the following columns:
 *                      INSURANCE_UID
 *
 *                      May also contain:
 *                      DEDUCTIBLE 
 *                      LIMIT
 *                      AGG_LIMIT
 *
 * COVERAGEDATA1 = Dataset with primary insurance coverage probabilities
 *                 This dataset has been geenrated by %orva_insurance_coverage   
 *                 and has possibly been altered by the application if 
 *                 coveage probabilities were changed by the user
 *
 * VERSION = Version of Data for losses that should be used.
 * 
 * CoverageProportion = Name of the dataset created that contains, primary coverage 
 *                      probabilities information along with the coverage proportions
 *                      calculated herein
 *                            
 * insStatus      = Name of status dataset indicating success of failure of 
 *                  the macro with additional information on the reason of the failure
 ******************************************************************************/
%macro orva_Insurance_Proportion(
                                LossData=,      
                                ccKey=,   
                                InsuranceData=,
                                CoverageData1=,
                                CoverageProportion=,
                                insStatus=insStatus,
                                version=   
                                );

    /*** set up the status macro variables  ***/
    %global status reason;
    %let status=0;
    %let reason=Successful;
    
    %let abort=0;

    /*** create the status data set upfront **/
    data &insStatus;
        format status 8. reason $100.;
        status=1;
        reason="Unknown";
        count=0;
    run;

    /* Check for the no internal data case ---------*/
    %orvl_ds_exists(&LossData, "&LossData", false);
    %if &ds_nobs eq 0 %then %do;
         %orvl_ds_exists(VARPRJ.CC, "VARPRJ.CC", false);
         %if &ds_nobs eq 0 %then %do; 
             %let cc_ds = VARPRJ_M.CC;
         %end;
         %else %do;
             %let cc_ds = VARPRJ.CC;
         %end;

         data &CoverageProportion; set &cc_ds(keep=ccuid);
            insurance_1 = 0;
            PIns1 = 1;
            PInsCurrent1 = 1;
            CoverageProp = 0;
         run;
         %goto DONE;
    %end;
    

    /* Replacing missing data by zeroes in the Loss dataset*/
    data _lossX (compress=binary);
        set &LossData;
        keep &cckey Insurance_1 Insurance_2 Insurance_Recovery_1 Insurance_Recovery_2;

        if Insurance_1=. then do;
            Insurance_1=0;
        end;

        if Insurance_2=. then do;
            Insurance_2=0;
        end;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to replace missing data by zeroes in the &LossData dataset.;
        %let abort=1;
        %let reason=Unable to replace missing data by zeroes in the &LossData dataset;
        %goto DONE;
    %end;

    /* Extend the insurance dataset by adding a row with a zero Insurance_UID*/
    DATA _insuranceX;
        set &InsuranceData nobs=count; where version_UID = &version;
 		keep INSURANCE_UID DEDUCTIBLE LIMIT AGG_LIMIT;
        output;
        
        if _N_=count then do;
            INSURANCE_UID                 = 0;
            DEDUCTIBLE                    = .;
            LIMIT                         = .;
            AGG_LIMIT                     = .;
 			output;
        end;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to add a row with a zero Insurance_UID.;
        %let abort=1;
        %let reason=Unable to add a row with a zero Insurance_UID;
        %goto DONE;
    %end;

    /*Reproduce the extended insurance dataset with different names for secondary coverage*/
    data _insurance2(rename=(INSURANCE_UID=INSURANCE_UID2 Deductible=Deductible2 Limit=Limit2 AGG_Limit=AGG_Limit2)) ;
        set _insuranceX;    
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to create insurance data for secondary coverage.;
        %let abort=1;
        %let reason=Unable to create insurance data for secondary coverage.;
        %goto DONE;
    %end;

    /*Denormalize data by combining the losses with the insurance information*/
    proc sql;
        create table _lossTmp 
        as
        select * from 
                _lossX a left join _insuranceX b 
        on  
                a.Insurance_1 = b.Insurance_UID
        left join _insurance2 c 
                on 
                a.Insurance_2 = c.Insurance_UID2
        ;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to denormalize insurance data.;
        %let abort=1;
        %let reason=Unable to denormalize insurance data;
        %goto DONE;
    %end;

    /* Sort the denormalize data by ccuid, Insurance_1 and Insurance_2 */
    proc sort data=_lossTmp;
        by  &cckey 
            Insurance_1
            insurance_2
            ;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to sort the denormalize data. ; 
        %let abort=1;
        %let reason=Unable to sort the denormalize data.;
        %goto DONE;
    %end;


    /* Find the number of covered losses for each type of coverage*/
    proc summary data=_lossTmp;
        by  &cckey 
            Insurance_1 
            Insurance_2
            ;
        var Insurance_Recovery_1 Insurance_Recovery_2;
        output  
                Sum=Sum_1 Sum_2 
                N=Count_1 Count_2 
                out=_coverage
                ;

    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to find the number of covered losses for each type of coverage.; 
        %let abort=1;
        %let reason=Unable to find the number of covered losses for each type of coverage;
        %goto DONE;
    %end;


    /* Make a copy of _coverage, call it _coverage_1*/
    data _coverage_1;
        set _coverage;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to make a copy of _coverage.; 
        %let abort=1;
        %let reason=Unable to make a copy of _coverage;
        %goto DONE;
    %end;

    /*    Make a copy of _coverage_1 to merge with*/
    data _coverage_2(rename=(Insurance_2=Insurance_1));
        set _coverage_1(drop=Insurance_1 Sum_1 Count_1);
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to make a copy of _coverage_1.; 
        %let abort=1;
        %let reason=Unable to make a copy of _coverage_1;
        %goto DONE;
    %end;

    /*    Sort by CCUID and Insurance_1    */
    proc sort data=_coverage_2;
        by 
            &cckey 
            Insurance_1;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to complete sort by &cckey and Insurance_1.; 
        %let abort=1;
        %let reason=Unable to complete sort by &cckey and Insurance_1;
        %goto DONE;
    %end;

    /* Take sum of secondary coverage */
    proc summary data=_coverage_2;
        by 
            &cckey 
            Insurance_1;
        var Sum_2;
        output
            SUM=TotalSum_2
            out= _Summ_2_;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to find sum of secondary coverage.; 
        %let abort=1;
        %let reason=Unable to find sum of secondary coverage;
        %goto DONE;
    %end;

    /* Sort  */
    proc sort data=_coverage_1;
        by 
            &cckey 
            Insurance_1;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to complete sort by &cckey and Insurance_1.; 
        %let abort=1;
        %let reason=Unable to complete sort by &cckey and Insurance_1;
        %goto DONE;
    %end;
        
    /* Take sum of primary coverage */ 
    proc summary data=_coverage_1;
        by 
            &cckey 
            Insurance_1;
        var Sum_1;
        output
            SUM=TotalSum_1
            out= _Summ_1_;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to find sum of primary coverage.; 
        %let abort=1;
        %let reason=Unable to find sum of primary coverage;
        %goto DONE;
    %end;

    /*    Do the merge*/
    data _coverage_3;
        merge _Summ_1_  _Summ_2_;
        by 
            &cckey 
            Insurance_1;

        if totalSum_1 ne . then do;
            if totalsum_2 ne . then do;
                TotalSum=TotalSum_1+TotalSum_2;
            end;
            else do;
                    TotalSum=TotalSum_1;
            end;
        end;
        else do;
            if totalsum_2 ne . then do;
                TotalSum = TotalSum_2;
            end;
            else do;
                TotalSum=0;          
            end;
        end;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to find sum of primary coverage.; 
        %let abort=1;
        %let reason=Unable to find sum of primary coverage;
        %goto DONE;
    %end;

    /* Do the sort */
    proc sort data=_coverage_3;
    by Insurance_1;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to complete sort.; 
        %let abort=1;
        %let reason=Unable to complete sort;
        %goto DONE;
    %end;

    /* Calculate the total recovery per contract */
    proc summary data= _coverage_3;
    by Insurance_1;
    var TotalSum;
    output
            sum=TotalTotal
            out=_coverage_4
            ;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to obtain the total insurance recovery per policy.; 
        %let abort=1;
        %let reason=Unable to obtain the total insurance recovery per policy;
        %goto DONE;
    %end;

    /* Merge the total amount of recovery per policy
       with the total amount of recovery per CCUID
       and obtain the coverage proportion */
    data _coverage_5(keep= Insurance_1 CCUID CoverageProp);
    merge   _coverage_3(keep= Insurance_1 CCUID TotalSum)
            _coverage_4(keep= Insurance_1 TotalTotal);
    by Insurance_1;
    
    if  Insurance_1=0 then do;
            CoverageProp=0;
    end;
    else do; 
            if totalTotal>0 then do;
                CoverageProp=TotalSum/totalTotal;
            end;
            else do;
                CoverageProp=0;
            end;
        end;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to complete the merge to obtain the coverage proportions.; 
        %let abort=1;
        %let reason=Unable to complete the merge to obtain the coverage proportions;
        %goto DONE;
    %end;

    /* Sort by CCUID and by policy */
    proc sort data=_coverage_5;
    by &cckey Insurance_1;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to complete sort.; 
        %let abort=1;
        %let reason = Unable to complete sort;     
        %goto DONE;
    %end;
    

    /* Compute proportions of primary coverage*/
    data &CoverageProportion 
                        (keep = &cckey 
                                Insurance_1 
                                PIns1 
                                PInsCurrent1 
                                CoverageProp
                        );
        merge &CoverageData1 _coverage_5;
        by &cckey Insurance_1;
        if PIns1 EQ . then do;
            PIns1=0;
            PInsCurrent1=0;
        end;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to combine primary coverage with overall coverage probabilities.; 
        %let abort=1;
        %let reason=Unable to combine primary coverage with overall coverage probabilities;
        %goto DONE;
    %end;


%DONE:

    %if &abort %then %do;
        %put ERROR: orva_insurance_proportion macro ended abnormally.;
        %let status=1;
    %end;

    /*- delete unwanted data -*/
    proc datasets lib=work memtype=data nolist;
        delete  _Summ_:
                _coverage 
                _coverage_: 
                Final_tmp 
                _loss:
                _insurance:
                ;
    run;
    quit;

    data &insStatus;
        format status 8. reason $100.;
        status=&status;
        reason="&reason";
    run;
%mend orva_Insurance_Proportion;

