 /*****************************************************************************************
 * Copyright (C), 2005
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   NAME:    ORVA_Loss_Sim.sas                                            
 *   PURPOSE: Simulate aggregate loss from frequency and severity distributions 
 *            This IS THE FASTER MC SIMULATION VERSION (b)!!!! 
 *   SUPPORT: jariou            
 *
 *
 * PARAMETERS:
 *
 * SPARMPRIORITYDATA = the name of the priority parameter data set with the following variables:
 *                     dataset_name...
 *
 * COVERAGEDATA1 = The name of the dataset containing primary
 *                 insurance coverage probabilities.
 *
 * COVERAGEDATA2 = The name of the dataset containing secondary
 *                 insurance coverage probabilities.
 *
 * INSURANCEDATA = The name of the dataset containing the insurance accounts.
 *
 * VERSION       = current data version
 *
 * LOSSKRIDATA   =  Dataset containing losses with KRI info
 *
 * KRIWEIGHTS =  Dataset containg ccUID, BL_UID, and the proportions
 *                  of losses of each BL_UID by CCUID
 *
 * freqSimData   = The data set containing the frequency draws from orva_freq_sim.sas (required in 3.2)
 *
 * SIMDATE       = The simulation date.
 *
 * SCALEDATE  = The date all scaling is done to (ALWAYS = global end date of internal data)
 *
 * CCUID         =  The cell number for the simulation ( required in 3.2 )
 *
 * seedData      =  dataset of seeds for each cell CCUID.
 *
 * ConfIntData = The output file from the orva_cisetup macro.  Contains the indices into the 
 *             order stats for the percentile point estimates and conf intervals.
 *
 * outSimNaked   =  the name of the replication output data set with the following variables:
 *                    _REP_ and LOSS. (W/O insurance)
 *
 * outsiminsured  =  the name of the replication output data set. (W/ insurance)
 *
 * OUTSTATNaked  =  the name of the output statistics data set with the following variables:
 *                  STATUS, REASON, FDIST, SDIST, MEAN, NPCTL, PCTL1-PCTLn, P_1-P_n,
 *                  LCL1-LCLn, UCL1-UCLn, and COVERAGE1-COVERAGn. (W/O insurance)
 *
 * OUTSTATInsured   =  the name of the output statistics data set. (W/ insurance)
 *
 *  parallel        = 1: use parallel processing
 *                    0: don't use parallel processing
 *
 ***********************************************************************************/
/*                  Other helper macros in this file                                     */
/*                                                                                       */
/*    %macro compute_expected_shortfall( simdata =, outstat =,  nreps= );                */
/*    %macro HandleBucketScenario(baseData=, counter=);                            */
/*    %macro ApplyInsurance(Level=);                                               */
/*    %macro HandleModelScenario(baseData=,counter=);                              */
/*    %macro HandleModelCase(DistNameVar=, ScaleEffect=);                          */
/*    %macro HandleKRIs();                                                         */
/*    %macro SetUpInsuranceMacros(CoverageData1  = ,                               */
/*                            CoverageData2  = ,                                   */
/*                            InsuranceData = ,                                    */
/*                            Version       = ,                                    */
/*                            CCUID= );                                            */
/*    %macro CreateMainLookupTable;                                                */
/*                                                                                       */
/***********************************************************************************/
/*         Main list of macro variables created and/or used in this file           */ 
/*---------------------------------------------------------------------------------*/
/* Source Types:  I   Internal, i.e. model parameters                               */
/*                                                                                  */
/*                F   Fit Scenario, i.e. model parameters                           */
/*                M   Mixed, i.e. model parameters                                  */
/*                N   Mixed Scenario, i.e model parameters                          */
/*                P   Public, i.e. model parameters                                 */
/*                C   Consortium, i.e. model parameters                             */
/*                                                                                  */
/*                S   Scenario proxy                                                */
/*                R   Rare Event Scenario                                           */
/*                                                                                  */
/*---------------------------------------------------------------------------------*/
/*   Scenario Type:  M   Model Based  Senario                                       */
/*                   B   Buckect Based Scenario                                     */
/*---------------------------------------------------------------------------------*/
/*                      List of macro variables available                           */
/*---------------------------------------------------------------------------------*/
/*                                                                                  */
/*    sParmPriorityData = Dataset with list of priorities for severity simulation   */
/*      SOURCE_TYPE1-nMixIntervals               =  Source_Type                     */
/*      SOURCE_NAME1-nMixIntervals               =  Source_Name                     */
/*      DATASET_NAME1-nMixIntervals              =  Dataset_name                    */
/*      SECONDARY_DATASET_NAME1-nMixIntervals    =  Secondary Dataset Name          */
/*      SCENARIO_TYPE1-nMixIntervals             =  Scenario Type                   */
/*      LOWER_BOUND1-nMixIntervals               =  Lower Bound                     */
/*      UPPER_BOUND1-nMixIntervals               =  Upper Bound                     */
/*---------------------------------------------------------------------------------*/
/*                                                                                  */
/*    nReps             = Number of simulations to run                              */
/*    seedData          = Name of seed dataset                                     */
/*    seed              = The seed value for this run                               */
/*    outStatInsured    = Dataset name of simulation with insurance applied         */
/*    sevParmData       = Dataset name of severity parameters                       */
/*    MAXLOSS           = Maximum simulated loss amount for this cell               */
/*    nMixIntervals     = Number of mixing intervals in the simulation              */
/*    nRareEventScen    = Number of rare event scenarios                            */
/*---------------------------------------------------------------------------------*/

%macro ORVA_Loss_Sim(   sParmPriorityData  =  ,  
                        coverageData1   =  ,   /* applyInsurance=??? */
                        coverageData2   =  ,   /* applyInsurance=??? */
                        insuranceData   =  ,
                        version         =  ,
                        lossKriData     =  ,
                        freqSimData     =  ,
                        ccUid           =  ,
                        seedData        =  ,
                        lookupSize      =  1000,
                        outsimNaked     =  outsimNaked,     
                        outsimInsured   =  outsimInsured,
                        outstatNaked    =  outstatNaked,   
                        outstatInsured  =  outstatInsured,
                        ConfIntData     =  ,
                        parallel        =  0,
                        nstreams        =  1,
                        simDate         =  ,
                        scaleDate       =  ,
                        cpiAdjustment   =  1.0
                    );

   %global abort;
   %global PARAM_COUNT;
   %global pInsuranceCount;
   %global TOTAL_SCEN_MEAN;

   %let KRI_ADJUST=0;
   %let abort=0;

   %let status=0;
   %let reason=Successful;
   %let fdist=;
   %let sdist=;
   %let PROXY_RUN_DONE = No;  
   %let AVERAGE_KRI_EFFECT = 1;
   %let UPPER_BOUNDS =%str();

   %if "&simDate" eq "" %then %do;
      %let simDate= %sysfunc(today());
   %end;
   %if "&scaleDate" eq "" %then %do;
      %let scaleDate= %sysfunc(today());
   %end;
   
    /* Put the simulation date into a numerical format */
    /* so it is easy to do calculations with           */
    data _NULL_;
        format SimDate BEST12.;
        SimDate=&simDate;
        call symput("simDate", SimDate);
    run;    

   /*- create the status data set upfront -*/
    data &outstatNaked;
        format status 8. reason $100.;
        status=1;
        reason="Unknown";
    run;

   data &outstatInsured;
        format status 8. reason $100.;
        status=1;
        reason="Unknown";
    run;

   /*- error checking -*/
    /* sParmPriorityData must be specified */
    %if ("&sParmPrioritydata" EQ "") %then %do;
      %put ERROR: The parameter SPARMPRIORITYDATA= must be specified.;
      %let abort=1;
      %let reason=MISSING SPARMPRIORITYDATA= PARAMETER;
      %goto DONE;
   %end;

/* Get the number of replications: BEGIN */ 
    /* Set the macro variable nreps to know how many simulations to run */
    data _null_;
        set &ConfIntData;
        call symput("nreps",nreps);
    run;
 
    /* The number of simulations must be positive */
   %if &nreps<=0 %then %do;
      %put ERROR: The number of simulations, NREPS, must be positive.;
      %let abort=1;
      %let reason=NONPOSITIVE REPLICATIONS;
      %goto DONE;
   %end;

    /* nreps must be an integer */
   %if %sysevalf(&nreps,integer) ^= &nreps %then %do;
      %put ERROR: The number of simulations, NREPS, must be a positive integer.;
      %let abort=1;
      %let reason=NON-INTEGER REPLICATIONS;
      %goto DONE;
   %end;
/* Get the number of replications: END */

/* Get the seed information: BEGIN */
    /* seedData must be specified */
    %if "&seedData" EQ "" %then %do;
      %put ERROR: Parameter seedData MUST be provided!;
      %let abort=1;
      %let reason=MISSING SEEDATA;
      %goto DONE;
   %end;
/* Get the seed information: END */
   
/* Frequency simulation: BEGIN */
    /* Obtain the name of the right column of frequencies */
    data _NULL_;
        CELL = compress('Cell_'|| put(&CCUid, &ORVAR_CELLFMT));
        call symput('_CELL', CELL);
      run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to obtain the name of the cell in the right format for extraction of frequencies.;
        %let abort=1;
        %let reason=Unable to obtain the name of the cell in the right format for extraction of frequencies;
        %goto DONE;    
    %end;

    /*  */
    data _loss_(keep=nloss);
        set &freqSimData;   
        /* There is a column in freqSimData for each cell in the VaR matrix */
        nloss = &_CELL;
        if nloss = . then do;
           call symput("abort", 1 );
           call symput("reason", "Simulated frequencies for &_CELL are missing." );
           stop;
        end;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to obtain the simulated frequencies.;
        %let abort=1;
        %let reason=Unable to obtain the simulated frequencies;
        %goto DONE;    
    %end;
    /* Frequency simulation: END */
    %if &abort=1 %then %GOTO DONE;

    /* Severity simulation preparation: START */
    /* It creates _SevParmsProxy_ */
    %CreateMainLookupTable;
    %if &abort=1 %then %GOTO DONE;

    %let nRareEventScen=0;
    data _SevParmsProxy_;
        set &sParmPriorityData; 
        WHERE CCUID = &ccUID AND SOURCE_TYPE EQ "R";
        call symputX("nRareEventScen",_N_);
    run;

   %let RARE_LOOKUP_QUANTILES = .;
 
    %if "&nRareEventScen" NE "0" %then %do;
        %let TOTAL_SCEN_MEAN = 0;
        %CreateRareEventLookupTable;
        %if &abort=1 %then %GOTO DONE;
    %end;
    /* Severity simulation preparation: END */
    
    /* Setting up insurance related macros: BEGIN */
    %if "&coverageData1" ne "" %then %do;
        %SetUpInsuranceMacros(
                            CoverageData1  = &CoverageData1,
                               CoverageData2  = &CoverageData2,
                               InsuranceData  = &InsuranceData,
                               Version        = &Version,
                            CCUID         = &CCUID 
                            );
        %if &abort=1 %then %GOTO DONE;
    %end;
    /* Setting up insurance related macros: END */

  
    /*- Monte Carlo simulation -*/
    /************************ Parallelized loop ***************************/

    data _NULL_;
       now = datetime();
       call symput('_TimeStart_',now);
    run;

    %global ORVP_STATUS;
    %let ORVP_STATUS= 1;
    %if ("&parallel" eq "1") %then %do;
        %orvp_distribute_loss_sim(nreps=&nreps,nstreams=&nstreams,seedData=&seedData)
        %if (&ORVP_STATUS eq 0) %then %do;
            data &outsimNaked;
                set _osimNaked_;
            run;
            data &outsimInsured;
                set _osimInsured_;
            run;
        %end;
        %else %do;
            %put WARNING: No processors available to perform parallel simulation. Reverting to sequential simulation.;
        %end;
    %end;
    %if (&ORVP_STATUS ^= 0) %then %do;
        data _NULL_;
           reps = ceil(&nreps/&nstreams);
           call symput('REPS', reps);
        run;
        %let NITER= 1;
        %let STARTREP= 1;
        %let NREPSLEFT= &nreps;
        %do %while (&NREPSLEFT gt 0);

           %let SEED = 0;
           data _NULL_; 
             set &seedData; where CCUID eq &ccUID and HOST eq &NITER;
             length tseed $ 32;
             tseed = put(seed,32.);
             call symput("SEED", trim(left(tseed)));
           run;

           %orva_loss_sim_stream(startRep=&STARTREP,nreps=&REPS,streamNum=&NITER,seed=&SEED);

           data &outsimNaked;
                set %if &NITER ne 1 %then %do; &outsimNaked %end;
                    _osimNaked_&NITER;
           run;
           data &outsimInsured;
                set %if &NITER ne 1 %then %do; &outsimInsured %end; 
                    _osimInsured_&NITER;
           run;
           proc datasets library=work memtype=data nolist nodetails;
              delete _osimNaked_&NITER _osimInsured_&NITER;
           run;
            
           %let STARTREP= %eval(&STARTREP + &REPS);
           %let NREPSLEFT= %eval(&NREPSLEFT - &REPS);
           %if &NREPSLEFT lt &REPS %then %do;
               %let REPS = &NREPSLEFT;
           %end;
           %let NITER=%eval(&NITER+1);
        %end;
    %end;

    data _NULL_;
       now = datetime();
       elapsed = now-symget('_TimeStart_');
       call symput('_TimeElapsed_',elapsed);
    run;
    %put ORVA_Loss_Sim Elapsed Time: &_TimeElapsed_;

    /************************ Parallelized loop ***************************/


    %orva_get_var_stats(
                        simData     = &outsimNaked, 
                        outStat     = &outStatNaked, 
                        lossVar     = loss,
                        confIntData = &confIntData
                        );   
    %compute_expected_shortfall( 
                                simData = &outsimNaked, 
                                outstat = &outstatNaked, 
                                nreps   = &nreps 
                                );


        %orva_get_var_stats(
                            simData     = &outsimInsured, 
                            outStat     = &outstatInsured, 
                            lossVar     = loss,
                            confIntData = &confIntData,
                            CoverageData= &CoverageData1                          
                            );   
    %compute_expected_shortfall( simData = &outsimInsured, 
                                 outstat = &outstatInsured, 
                                 nreps   = &nreps 
                                );




%DONE:

    %if &abort %then %do;
        %put ERROR: LossSim macro ended abnormally.;

        data &outstatNaked;
            format status 8. reason $100.;
            status=1;
            reason="&reason";
        run;
    %end;    


   /*- delete unwanted data sets -*/
   proc datasets library=work memtype=data nolist nodetails;
      delete 
          _temp0_
          _temp_
         _loss_ 
         _TParm_ 
         _tmp_kri_
         _tmp_kri_summ_
        ;
   run;
   quit;

%mend ORVA_Loss_Sim;


/* Compute the tail index and a second set of VaR numbers */
/* based on the expected shortfall -----------------------*/
/* Outstat should contain the var statistics computed thus*/
/* far  --------------------------------------------------*/
%macro Compute_Expected_Shortfall( simdata =, outstat =,  nreps= );

    /* Compute the tail index from the 99.2 point -----------*/
     data _NULL_; set &outstat;
        call symput("Tail_Point", P_3); /* default if < 1000 draws */
        call symput("median", median);
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to set TailPoint and Median macros.;
        %let abort=1;
        %let reason=Unable to set TailPoint and Median macros;
        %goto SHORTFALL_DONE;    
    %end;

    proc sort data=&simdata out=tmp(keep=Loss); 
    by loss; run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to sort the simulated data.;
        %let abort=1;
        %let reason=Unable to sort the simulated data;
        %goto SHORTFALL_DONE;    
    %end;

    %let Tail_Point=.;

    data _NULL_; 
        if &nreps >= 1000 then do;
           n= ceil(&nreps*0.992);
           set tmp point=n;
           call symput("Tail_Point", loss);
        end;
        stop;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to set TailPoint macro.;
        %let abort=1;
        %let reason=Unable to set TailPoint macro;
        %goto SHORTFALL_DONE;    
    %end;

    %let alpha = .;
    %let M     = .;

    data _NULL_; 
        set &simdata end=last;
        where loss > &Tail_Point;
        cnt+1;
        invalpha + log( (loss-&median)/ (&Tail_Point-&median));

        if last then do;
            alpha = cnt / invalpha;
            call symput("alpha", alpha);
            call symput("M", cnt);
        end;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to obtain tail index.;
        %let abort=1;
        %let reason=Unable to obtain tail index;
        %goto SHORTFALL_DONE;    
    %end;

    data &outstat; 
        set &outstat;

        
        label TailIndex = "Tail index";
        TailIndex = &alpha;

        label VaREVT_1 = "VaR computed using Extreme Value Theory for percentile 1";
        VaREVT_1 = (&Tail_Point -median) * ( &m/ &nreps /(1- pctl1/ 100) )**(1/&alpha) + median;

        label VaREVT_2 = "VaR computed using Extreme Value Theory for percentile 2";
        VaREVT_2 = (&Tail_Point -median) * ( &m/ &nreps /(1- pctl2/ 100) )**(1/&alpha) + median;

        label VaREVT_3 = "VaR computed using Extreme Value Theory for percentile 3";
        VaREVT_3 = (&Tail_Point - median) * ( &m/ &nreps /(1-pctl3/ 100))**(1/&alpha) + median;

        label CondExpectedLoss_1 = "Conditional expected loss for percentile 1";
        CondExpectedLoss_1 = VaREVT_1 * &alpha /(&alpha-1); 
        if CondExpectedLoss_1 < 0 then CondExpectedLoss_1 = .;

        label CondExpectedLoss_2 = "Conditional expected loss for percentile 2";
        CondExpectedLoss_2 = VaREVT_2 * &alpha /(&alpha-1); 
        if CondExpectedLoss_2 < 0 then CondExpectedLoss_2 = .;

        label CondExpectedLoss_3 = "Conditional expected loss for percentile 3";
        CondExpectedLoss_3 = VaREVT_3 * &alpha /(&alpha-1);
        if CondExpectedLoss_3 < 0 then CondExpectedLoss_3 = .;

    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to write out expected shortfall output.;
        %let abort=1;
        %let reason=Unable to write out expected shortfall output;
        %goto SHORTFALL_DONE;    
    %end;

%SHORTFALL_DONE:
%mend Compute_Expected_Shortfall;

%macro HandleBucketScenario(baseData=, counter=);
    data _null_;
        set &DSNAME;
        where CCUID=&CCUID;
        call symput("NBUCKETS", NBUCKETS);
    run;

    
    /* All bucket scenarios are merged together at this point? */
    data _DS_CCUID_;
        set &DSNAME;
        Where CCUID=&CCUID;
        Threshold=0;
    run;

    data _temp_;
        set &baseData;
        call symput("effLookupSize",_N_);
    run;

    data _lookup&counter.TMP;
        merge _temp_ _DS_CCUID_; 
        by threshold; 
    run;

    data _lookup&counter(KEEP= ecdf quant&counter);
        set _lookup&counter.TMP;
        retain  quant&counter 0  BucketIndex 1 lastP 0 bucketProb 0 lastCdf 0 lastQ 0 ;

        array final_bucketwgt[&nbuckets];
        array final_bucket[&nbuckets];

        /* This has been changed because later on, it is assumed */
        /* that this has been sampled equiprobably               */
         ecdf = 1-(1-_N_/(&effLookupSize+1))**2.5;
		/* ecdf = _N_/(&effLookupSize+1);*/

        if _N_ = 1 then do; 
            bucketProb = final_bucketwgt[1];
            lastQ=final_bucket0;
        end;

        if ecdf > bucketProb then do;
            quant&counter = (ecdf-lastCdf)/final_bucketwgt[bucketIndex] + lastQ;
            bucketIndex   = BucketIndex + 1; 
            lastCdf       = bucketProb;
            lastQ         = quant&counter;
            bucketProb    = bucketProb +  final_bucketwgt[BucketIndex];
        end;

        if BucketIndex =1 then do;
            quant&counter = (final_bucket[bucketIndex] - final_bucket0)*(ecdf-lastCdf)/final_bucketwgt[bucketIndex] + lastQ;  
        end;
        else do;
            quant&counter = (final_bucket[BucketIndex]- final_bucket[BucketIndex-1]) *(ecdf-lastCdf)/final_bucketwgt[bucketIndex] + lastQ;  
        end;

        lastQ      = quant&counter;
        lastCdf    = ecdf;
    run;

%mend HandleBucketScenario;

/* 
    Original method to create a lookup table for simulation. 
    Since the quantile of the mixture of the various model 
    scenarios would be a mess to obtain, I simply simulate 
    the mixture itself and order the results as approximate 
    appropriate quantiles
*/
%macro HandleModelScenario(baseData=,counter=);
    data _NULL_;
        set &baseData;
        call symputX("effLookupSize",_N_);
    run;

    /* obtain the parameter dataset for that cell */
    data _DS_CCUID_;
        set &DSNAME;
        where CCUID=&CCUID;
        call symput("modelCount",_N_);
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to obtain the parameter dataset for cell with CCUID=&CCUID.;
        %let abort=1;
        %let reason = Unable to obtain the parameter dataset for cell with CCUID=&CCUID;
        %goto DONE;    
    %end;
    
    /* If only one line, then treat this as an internal parametric fit and get out of here*/
    %if &modelCount=1 %then %do;
        %HandleModelCase(DistNameVar=Severity_Dist, ScaleEffect=1);
        %goto DONE_MODEL_SCEN;
    %end;

    /* If we are here, we know that there are more than one line. */
    /* Get the frequency mean for each model scenario, as well as */
    /* the aggregated mean frequency                              */
    data _DS_CCUID_;
        set _DS_CCUID_;
        retain TotalFreqMean 0;

        if Frequency_Dist = "POIS" then do;
            FreqMean = FParam_1;
        end;
        else do;
            FreqMean = FParam_1 * FParam_2;
        end;

        TotalFreqMean = TotalFreqMean + FreqMean;
        call symputX("TotalFreqMean",TotalFreqMean);
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to compute the frequency mean for each model scenario.;
        %let abort=1;
        %let reason = Unable to compute the frequency mean for each model scenario;
        %goto DONE;    
    %end;
    
    /* Order by value of frequency mean */
    proc sort data=_DS_CCUID_;
        by FreqMean;
    run;
    
    /* Find out how many from each scenario needs to be simulated                          */
    /* I floor the proportion and then make sure everything adds up to the desired count   */
    /* If does not add up, I add the missing count to the one with the largest proportion */
    data _DS_CCUID_;
        set _DS_CCUID_ end=Last;
        retain TotalCount 0;
        subCount = floor(FreqMean * 20*&lookupSize /&TotalFreqMean);
        TotalCOunt = TotalCOunt + subCOunt;

        if last then do;
            SubCount = SubCount + (20*&lookupSize - TotalCount);
            TotalCount = &lookupSize;
        end;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to compute the number of representatives from each scenario.;
        %let abort=1;
        %let reason = Unable to compute the number of representatives from each scenario.;
        %goto DONE;    
    %end;

    %do modelIndex = 1 %to &modelCount;
        data tmpSParm;
            set _DS_CCUID_;
            IF _N_= &modelIndex then do;
                output;
                call symput("Distribution",Severity_Dist);
            end;
        run;


        %orvd_&Distribution;

        %if &modelIndex = 1 %then %do;    
            data _lookupModelScen(KEEP= quant&counter);
                set tmpSParm;

                call streaminit(12345*&modelIndex);

                do i=1 to SubCount;
                    %sim&Distribution(quant&counter);
                    output;
                end;
            run;
        %end;
        %else %do;
            data _tempQuant_(KEEP = quant&counter);
                set tmpSParm;

                call streaminit(12345*&modelIndex);

                do i=1 to SubCount;
                    %sim&Distribution(quant&counter);
                    output;
                end;
            run;
            data _lookupModelScen;
                set _lookupModelScen _tempQuant_;
            run;
        %end;
    %end; /* of loop over modelIndex */

    proc sort data=_lookupModelScen;
        by quant&counter;
    run;

    %CompactifyLookupArray( inData   = _lookupModelScen,
                            quantVar = quant&counter,
                            outData  = _lookup&counter
                            );

%DONE_MODEL_SCEN:
%mend HandleModelScenario;

/* This macro handles creating a lookup table for a mixing interval where the  */
/* severity is described as a distribution model (either fit or scenario)      */
/* The only parameter is the name of the variable holding theseverity          */
/* distribution name.                                                          */
%macro HandleModelCase(DistNameVar=, ScaleEffect=);
    /* Set the distribution name in the SDist macro */
    data _null_;
        set &DSNAME;
        where CCUID=&CCUID;
        call symput("SDIST", &DistNameVar);
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to set the SDist macro variable.;
        %let abort=1;
        %let reason=Unable to set the SDist macro variable;
        %goto DONE_HANDLE_MODEL;    
    %end;

    /* Initialize the severity distribution */
    %orvd_&SDIST;

    /* Create the macros Param_1, Param_2, .... needed for the quantile calculations below */
    data _null_;
        set &DSNAME;
        where CCUID=&CCUID;
        array Params Param:;
        length TheParams $300;

        TheParams="Param_1 " || Param_1; 

        %if %sysevalf(&KRI_ADJUST > 0, boolean) %then %do;
            %if "&SOURCE_TYPE" NE "S" %then %do;
					TheParams = "Param_1 " || Orig_Param_1;
            %end;
        %end;
 
        do nParams=2 to dim(Params);
            call symput(vname(Params(nParams)),Params(nParams));
            TheParams = trim(theParams) || " " || vname(Params(nParams)) || "  " || Params(nParams);
        end;

        %if %sysevalf(&KRI_ADJUST > 0, boolean) %then %do;
            %if "&SOURCE_TYPE" NE "S" %then %do;
				%if "&SOURCE_TYPE" NE "P" %then %do;
					%if "&SOURCE_TYPE" NE "C" %then %do;
    		            call symput("Param_1",Orig_Param_1);
					%end;
				%end;
            %end;
        %end;
		  %else %do;
            call symput("Param_1", Param_1);	
		  %end;

        call symput("TheParams", TheParams); 
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to obtain and set the values of the macros Param_1, Param_2,... .;
        %let abort=1;
        %let reason=Unable to obtain and set the values of the macros Param_1, Param_2,...;
        %goto DONE_HANDLE_MODEL;    
    %end;

    /* Check to see  if a sample proxy dataset has been created yet */
    /* If no, then do one                                           */
    %if "&PROXY_RUN_DONE" = "No" %then %do;
        data _temp0_;
            retain &TheParams;
            call streaminit(12345);

            do i=1 to 20*&lookupSize;
                threshold=0;
                %Sim&sdist(Quantile)
                output;
            end;
        run;

        proc sort data=_temp0_;
            by Quantile;
        run;

        %let PROXY_RUN_DONE = Yes;

        %CompactifyLookupArray( inData   = _temp0_,
                                quantVar = quantile,
                                outData  = _temp0_
                                );

    %end;

    /* Get a clean copy of the _temp0_ dataset */
    data _temp_;
        set _temp0_;
        retain threshold 0;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to create clean copy of dataset _temp_.;
        %let abort=1;
        %let reason=Unable to create clean copy of dataset _temp_;
        %goto DONE_HANDLE_MODEL;    
    %end;

    /* Run the quantile macro for the distribution */
    /* It expect a _temp_ dataset and writes to it */
    %Quant&SDist;

    /* Get rid of extranuous variables in _temp_ */
    data _Lookup&nInt(keep= ecdf quant&nInt);
        set _temp_(rename=(quant=quant&nINt));
        quant&nInt = Quant&nInt * &ScaleEffect;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to clean up the lookup table dataset.;
        %let abort=1;
        %let reason=Unable to clean up the lookup table dataset;
        %goto DONE_HANDLE_MODEL;    
    %end;

%DONE_HANDLE_MODEL:

%mend HandleModelCase;

%macro HandleKRIs();
    %global KRI_CHOICES;
    /* If KRI's then compute the KRI values for this cell */
    %if ("&lossKriData" ne "") %then %do;   
        /* Get a list of the kri variables */
        /* See if there are KRIs */

        /* Subset the severity parm data to the CCUID of the cell */
        data _SEVPARMTMP;
            set &SevParmData;
            where CCUID = &ccUID;
            call symputX("KRI_ADJUST",COUNT_KRI);
        run;
        %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
            %put ERROR: Unable to subset severity parameter dataset for cell with CCUID=&CCUID.;
            %let abort=1;
            %let reason=Unable to subset severity parameter dataset for cell with CCUID=&CCUID;
            %goto DONE_HANDLE_KRIs;    
        %end;
    
        /* Check if KRI's were used in fitting this cell */
        %if &KRI_ADJUST > 0 %then %do;
            %orva_get_kri_log(data=_SEVPARMTMP);

            /* Calculate the KRI_SCALE_EFFECT for each loss in the dataset */
            data _tmp_kri_(keep=KRI_SCALE_EFFECT BLUID);
                set &lossKriData end=last;
                retain averageEffect 0;
                where CCUID=&CCUID;
                KRI_SCALE_EFFECT=exp(&KRI_LOG);
                averageEffect = averageEffect + KRI_SCALE_EFFECT;

                if last then do;
                    call symput("AVERAGE_KRI_EFFECT",averageEffect/_N_);
                end;
            run;
            %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
                %put ERROR: Unable to calculate the scale effect of the KRIs.;
                %let abort=1;
                %let reason=Unable to calculate the scale effect of the KRIs;
                %goto DONE_HANDLE_KRIs;    
            %end;

                
        %let dsid = %sysfunc(open(_tmp_kri_));
        %if &dsid %then  %do;
              %let KRI_CHOICES =%sysfunc(attrn(&dsid,nobs));
              %let rc = %sysfunc(close(&dsid));
        %end;
        %else %do;
            %put ERROR: Unable to obtain the total number of KRI choices.;
            %let abort=1;
            %let reason=Unable to obtain the total number of KRI choices;
            %goto DONE_HANDLE_KRIs;    
        %end;
        %end;
        %else %do;
            /* No KRI's are used */
            %let KRI_LOG=0;
            %let KRI_SCALE_ARRAY = 1;
            %let KRI_PROB_ARRAY= 1;
            %let AVERAGE_KRI_EFFECT=1;
        %end;
    %end;
    %else %do;
        %let KRI_ADJUST = 0;
        %let AVERAGE_KRI_EFFECT=1;
    %end; /* END OF "If there are KRI's */

%DONE_HANDLE_KRIs:

%mend; /* End of HandleKRIs */

%macro SetUpInsuranceMacros(CoverageData1  = ,
                            CoverageData2  = ,
                            InsuranceData = ,
                            Version       = ,
                            CCUID= );
    %GLOBAL CoverageCount     ;
    %GLOBAL pInsurance        ;
    %GLOBAL Insurance_UID_1   ;
    %GLOBAL Insurance_Index_1 ;
    %GLOBAL Deductible_1      ;
    %GLOBAL Limit_1           ;
    %GLOBAL Eff_Agg_Limit_1   ;
    %GLOBAL Payout_Haircut_1  ;
    %GLOBAL PolicyExpiryDate_1     ;
    %GLOBAL Insurance_UID_2   ;
    %GLOBAL Insurance_Index_2 ;
    %GLOBAL Deductible_2      ;
    %GLOBAL Limit_2           ;
    %GLOBAL Eff_Agg_Limit_2   ;
    %GLOBAL Payout_Haircut_2  ;
    %GLOBAL PolicyExpiryDate_2     ;
    %GLOBAL InsuranceContractCount ;
    %GLOBAL InsuranceArray         ;

    %let InsuranceContractCount = 0;
    data _tmp_InsuranceData_;
        length UID $8;
        set &InsuranceData;
        where Version_UID=&Version;
        UID=Insurance_UID;
        call symput("InsuranceContractCount",_N_);
        Insurance_Index=_N_;
    run;
    %if &InsuranceContractCount eq 0 %then %do;
       %let CoverageCount=0;
       %goto DONE_INS_SETUP;
    %end;
    
    proc SQL noprint;
        SELECT
            "Insurance_" || compress(UID)
        INTO
            :InsuranceArray SEPARATED BY " "
        FROM
             _tmp_InsuranceData_;
            ;
    quit;

    /*************************************************************/ 
    /* Compute aggregate probability of coverage                 */
    /* i.e. product of coverage by insurance 1 and insurance 2   */
    /* Thoroughly debugged... I think! ;-)                       */
    /*************************************************************/ 
    %let CoverageCount = 0;

    /*----- come up with CoverageProp1 and CoverageProp2 -----*/
    proc sort data=&coverageData1 out=_ins1; where CCUID=&CCUID; by CCUID Insurance_1; run;
    proc sort data=&coverageData2 out=_ins2; where CCUID=&CCUID; by CCUID Insurance_1; run;

    data _AllcoverageDatar;
        merge _ins1 _ins2;
        by CCUID Insurance_1;
        if PInsCurrent2 EQ . then do;
            Insurance_2=0;
            PInsCurrent2=1;
        end; 
        PInsurance = PInsCurrent1 * PInsCurrent2;
		  if missing(PInsurance) then PInsurance = 0;
        keep Insurance_1 Insurance_2 PInsurance CoverageProp;
        call symputx("CoverageCount",_N_);
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to compute aggregate probability of coverage.;
        %let abort=1;
        %let reason=Unable to compute aggregate probability of coverage;
        %goto DONE_INS_SETUP;    
    %end;

    data _AllcoverageDatar;
        if _N_ eq 1 then do;
            declare hash insHash(dataset: '_AllcoverageDatar', hashexp: 20);
            rc = insHash.defineKey('Insurance_1');
            rc = insHash.defineData('CoverageProp');
            rc = insHash.defineDone();
        end;
        set _AllcoverageDatar;
        CoverageProp1 = CoverageProp;
        save_Insurance_1 = Insurance_1;
        Insurance_1 = Insurance_2;
        CoverageProp2 = insHash.find();
        Insurance_1 = save_Insurance_1;
        drop CoverageProp save_Insurance_1;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to compute coverage proportion.;
        %let abort=1;
        %let reason=Unable to compute coverage  proportion.;
        %goto DONE_INS_SETUP;    
    %end;

    /*----- Add the insurance details to the coverage dataset -----*/
    data _AllcoverageDatar(keep =   Insurance_1 
                                    Insurance_2 
                                    PInsurance 
                                    Insurance_Index_1
                                    Deductible_1 
                                    Limit_1 
                                    Eff_Agg_Limit_1 
                                    Agg_Limit_1 
                                    Payout_Haircut_1
                                    PolicyExpiryDate_1
                                    CoverageProp1
                                    CoverageProp2
                                    );

            merge _AllcoverageDatar(in=ini) 
                _tmp_InsuranceData_(rename=
                                    (
                                     insurance_uid   =  insurance_1 
                                     Insurance_index =  insurance_index_1
                                     Deductible      =  Deductible_1
                                     Limit           =  Limit_1
                                     Agg_Limit       =  Agg_Limit_1
                                     Payout_Haircut  =  Payout_Haircut_1
                                     Effective_To_Date =  PolicyExpiryDate_1
                                    )
                                );
           format PolicyExpiryDate_1 BEST12.;
           by insurance_1;
           if ini;
           Eff_Agg_Limit_1 = CoverageProp1 * Agg_Limit_1;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to merge policy information with coverage information.;
        %let abort=1;
        %let reason=Unable to merge policy information with coverage information;
        %goto DONE_INS_SETUP;    
    %end;

    /*----- Sort insurance data by the second level of insurance coverage -----*/
    proc sort data=_AllCoverageDatar;
        by insurance_2;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to sort insurance data.;
        %let abort=1;
        %let reason=Unable to sort insurance data;
        %goto DONE_INS_SETUP;    
    %end;

    data _AllcoverageDatar(keep =   Insurance_1 
                                    Insurance_2 
                                    PInsurance 
                                    Deductible_1 
                                    Limit_1 
                                    Eff_Agg_Limit_1 
                                    Agg_Limit_1  
                                    Payout_Haircut_1 
                                    Insurance_index_1
                                    PolicyExpiryDate_1
                                    Deductible_2 
                                    Limit_2 
                                    Eff_Agg_Limit_2 
                                    Agg_Limit_2 
                                    Payout_Haircut_2
                                    Insurance_index_2
                                    PolicyExpiryDate_2
                                    );
            merge _AllcoverageDatar (in=ini) 
                _tmp_InsuranceData_(rename=
                                        (
                                        insurance_uid     =  insurance_2 
                                        Insurance_Index   =  Insurance_index_2
                                        Deductible        =  Deductible_2
                                        Limit             =  Limit_2
                                        Agg_Limit         =  Agg_Limit_2
                                        Payout_Haircut    =  Payout_Haircut_2
                                        Effective_to_Date =  PolicyExpiryDate_2
                                        )
                                    );
               by insurance_2;
               format PolicyExpiryDate_2 BEST12.;

               if ini;
               Eff_Agg_Limit_2 = CoverageProp2 * Agg_Limit_2;
    run;

    /*----- Count the number of insurance policies -----*/
    data _NULL_;
        set _tmp_InsuranceData_ end=last;
        if last then do;
            call symputx("InsurancePolicyCount",_N_);
        end;
    run;

    /*---- Count the number of insurance probabilities ----*/
    data _NULL_;
        set _AllcoverageDatar end=last;
        if last then do;
            call symputx("pInsuranceCount",_N_);
        end;
    run;

    proc sql noprint;
        SELECT 
            pInsurance,
            Insurance_1,
            Insurance_Index_1,
            Deductible_1,
            Limit_1,
            Eff_Agg_Limit_1,
            Payout_Haircut_1,
            PolicyExpiryDate_1,
            Insurance_2,
            Insurance_Index_2,
            Deductible_2,
            Limit_2,
            Eff_Agg_Limit_2,
            Payout_Haircut_2,
            PolicyExpiryDate_2
        INTO 
            :pInsurance          SEPARATED BY ", " ,
            :Insurance_UID_1     SEPARATED BY " "  ,
            :Insurance_Index_1   SEPARATED BY " "  ,
            :Deductible_1        SEPARATED BY " "  , 
            :Limit_1             SEPARATED BY " "  ,
            :Eff_Agg_Limit_1     SEPARATED BY " "  ,
            :Payout_Haircut_1    SEPARATED BY " "  ,
            :PolicyExpiryDate_1  SEPARATED BY " "  ,
            :Insurance_UID_2     SEPARATED BY " "  ,
            :Insurance_Index_2   SEPARATED BY " "  ,
            :Deductible_2        SEPARATED BY " "  ,
            :Limit_2             SEPARATED BY " "  ,
            :Eff_Agg_Limit_2     SEPARATED BY " "  ,
            :Payout_Haircut_2    SEPARATED BY " "  ,
            :PolicyExpiryDate_2  SEPARATED BY " "  
        FROM 
            _AllcoverageDatar
        ;
    quit;

%DONE_INS_SETUP:

   proc datasets library=work memtype=data nolist nodetails;
      delete _ins: _AllcoverageDatar;
   run;
%mend; /* SetUpInsuranceMacros */


%macro CreateMainLookupTable;
%global LOOKUP_QUANTILES;

    /* Set the lookup table up */
    data _temp0_;
        do i=1 to &lookupSize;
            threshold=0;
            output;
        end;
    run;

    /* Extract the following from sParmPriorityData: maxloss  */
    data _SevParmsProxy_;
        set &sParmPriorityData ; 
        WHERE CCUID = &ccUID AND SOURCE_TYPE NE "R";
        retain MaximumLoss 0;

        MaximumLoss = max(MaximumLoss,max_loss);
		run;
      %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to obtain the magnitude of the maximum likely loss.;
        %let abort=1;
        %let reason=Unable to obtain the magnitude of the maximum likely loss;
        %goto DONE_LOOKUP_MAIN;    
     %end;
		

		data _NULL_;
			set _SevParmsProxy_ end =last;

         if last then do;
            /* If all max_loss are missing  */
            if MaximumLoss=0 then do;
					MaximumLoss=10000000000;
				end;

				call symput("MAXLOSS",MaximumLoss);
        end;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to set the value of the MAXLOSS macro.;
            %let abort=1;
        %let reason=Unable to set the value of the MAXLOSS macro;
            %goto DONE_LOOKUP_MAIN;    
    %end;

    /*    Compactify _SevParmsProxy_*/
    %CompactifyRanges(data=_SevParmsProxy_, out=_SevParmsProxy_, ccuid=&CCUID);
    
    /* Check how many mixing intervals there are */ 
    %let dsid = %sysfunc(open(_SevParmsProxy_));
    %if &dsid %then  %do;
          %let nMixIntervals =%sysfunc(attrn(&dsid,nobs));
          %let rc = %sysfunc(close(&dsid));
    %end;
    %else %do;
        %put ERROR: No data for this cell in &sParmPriorityData.;
        %let abort=1;
        %let reason=No data for this cell in &sParmPriorityData;
        %goto DONE_LOOKUP_MAIN;
    %end;

    
    /* Check to see whether KRI's are used at all in any range */
    %let KRI_USED = No;

    data _NULL_;
        set _SevParmsProxy_;
        where SOURCE_TYPE="I";
        call symput("KRI_USED","Yes");
        call symput("SevParmData", DATASET_NAME);
    run;

    %if "&KRI_USED"="Yes" %then %do;
        %HandleKRIs;
        %if "&abort"="1" %then %GOTO DONE_LOOKUP_MAIN;
    %end;

    /*  Loop through each simulation interval    */
    /*  and set the appropriate macros            */
    %do nInt=1 %to &nMixIntervals;

        /* Set the Severity Dataset information line per line */
        data _null_;
            Index=&nInt;
            set _SevParmsProxy_ point=Index;
            call symput( "DSNAME", DATASET_NAME);
            call symput( "SOURCE_TYPE"   ,  SOURCE_TYPE   );
            call symput( "SCENARIO_TYPE" ,  SCENARIO_TYPE );
            call symput( "LOWER_BOUND"   ,  LOWER_BOUND   );
            call symput( "UPPER_BOUND"   ,  UPPER_BOUND   );
            stop;
        run;
        %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
            %put ERROR: Unable to obtain detailed information for mixing interval &nInt.;
            %let abort=1;
            %let reason=Unable to obtain detailed information for mixing interval &nInt;
            %goto DONE_LOOKUP_MAIN;    
        %end;
        
        /* Create a string of upper bounds values */
        %let UPPER_BOUNDS = &UPPER_BOUNDS &UPPER_BOUND;

        /* Case where source is Internal Data */
        %if     "&SOURCE_TYPE" EQ "I" 
            or 
                "&SOURCE_TYPE" EQ "F" 
            or 
                "&SOURCE_TYPE" EQ "M" 
            or 
                "&SOURCE_TYPE" EQ "N" 
            or 
                "&SOURCE_TYPE" EQ "C" 
            or 
                "&SOURCE_TYPE" EQ "P" 
                                        %then %do;

            %if "&SOURCE_TYPE" EQ "I" %then %do; 
                %let SevParmData = &DSNAME;

                /* I want to call the HandleKRIs macro only once */ 
                %if "&HandledKRIs" EQ "No" %then %do;

                    %HandleKRIs;
                    %if "&abort"="1" %then %GOTO DONE_LOOKUP_MAIN;
                    %let HandledKRIs=Yes;
                %end;
            %end;

            %HandleModelCase(DistNameVar=Distribution, ScaleEffect=&AVERAGE_KRI_EFFECT);
            %if "&abort"="1" %then %GOTO DONE_LOOKUP_MAIN;
        %end;
        /* Case where source is scenario */
        %else %if 
                "&SOURCE_TYPE" EQ "S" 
            %then %do;
            /*Case where the scenario is of Model type */
            %if 
                "&SCENARIO_TYPE" EQ "M" 
                %then %do;
                %HandleModelScenario(baseData=_temp0_, counter = &nInt); 
                %if "&abort"="1" %then %GOTO DONE_LOOKUP_MAIN;
            %end;
            /* Case where the scenario is of type Bucketed */
            %else %do;
                %HandleBucketScenario(baseData=_temp0_, counter = &nInt);
                %if "&abort"="1" %then %GOTO DONE_LOOKUP_MAIN;
            %end;
        %end;
        
        /* Create the lookup table for this mixing interval  */
        /* and merge with other mixing intervals             */
        %if "&nInt" = "1" %then %do;
            data _lookups_;
                set _lookup&nInt;
            run;
        %end;
        %else %do;
            data _lookups_;
                merge _lookups_ _lookup&nInt;
                by ecdf;
            run;
        %end;
        %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
            %put ERROR: Unable to create lookup table for mixing interval &nInt.;
            %let abort=1;
            %let reason=Unable to create lookup table for mixing interval &nInt;
            %goto DONE_LOOKUP_MAIN;    
        %end;
    %end; /* End of the interval looping */


    /* Create the file containing the array of the joined lookup tables */
    data _LookUp_Table_(keep= ecdf quantile);
        format quantile BEST12.;
        set _lookups_ end=last;
        array UB{&nMixIntervals}_TEMPORARY_ ( &UPPER_BOUNDS );
        array quant[&nMixIntervals];
        retain ArrayIndex 0 UpBound 0;

        if ecdf>=UpBound then do;
            ArrayIndex =  ArrayIndex + 1;
            UpBound = UB[ArrayIndex];
        end;
        
            quantile = 0;
        do i=1 to ArrayIndex;
            quantile = max(quantile, quant[i]);
        end;
    run;
    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to write out the lookup table as an array.;
        %let abort=1;
        %let reason=Unable to write out the lookup table as an array.;
        %goto DONE_LOOKUP_MAIN;    
    %end;


    /* Try this SQL to create the lookup table macro */
    proc SQL noprint;
        SELECT 
            quantile
        INTO
            :LOOKUP_QUANTILES SEPARATED by " "
        FROM
            _LookUp_Table_
        ;
    quit;

%DONE_LOOKUP_MAIN:
%mend CreateMainLookupTable;

/* Create a lookup table for rare events only. */
%macro CreateRareEventLookupTable;
    %do nScen=1 %to &nRareEventScen;
            /* Set the Severity Dataset information line per line */
            data _null_;
            Index=&nScen;
            set _SevParmsProxy_ point=Index;
                call symput( "DSNAME", DATASET_NAME);
                call symput( "SOURCE_TYPE"   ,  SOURCE_TYPE   );
                call symput( "SCENARIO_TYPE" ,  SCENARIO_TYPE );
                stop;
            run;
            %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
            %put ERROR: Unable to obtain detailed information for  scenario &nScen.;
                %let abort=1;
            %let reason = Unable to obtain detailed information for scenario &nScen;
                %goto DONE_LOOKUP_RARE;    
            %end;

        %if &nScen =1 %then %do;
            data _ScenMeans_ ;
                set &DSName end=last;
                where CCUID=&ccuid;
                %if "&SCENARIO_TYPE" = "M" %then %do;
                    retain ScenMean 0;

                    if Frequency_Dist = "POIS" then;
                        tmpScenMean = FParam_1;
        end;
        else do;
                        tmpScenMean = Fparam_1 * FParam_2;
        end;

                    ScenMean      = ScenMean + tmpScenMean;
                    DSName        = "&DSName";
                    Scenario_Type = "&Scenario_Type";
                    if last then output;
                %end;
                %else %do;
                    ScenMean      = Mean;
                    DSName        =  "&DSName";
                    Scenario_Type = "&Scenario_Type";
                %end;
    run;
    %end;
        %else %do;
            data _tmp_ScenMeans_;
                set &DSName;
                where CCUID=&ccuid;
                %if "&SCENARIO_TYPE" = "M" %then %do;
                    retain ScenMean 0;

                    if Frequency_Dist = "POIS" then;
                        tmpScenMean = FParam_1;
                    end;
                    else do;
                        tmpScenMean = Fparam_1 * FParam_2;
                    end; 

                    ScenMean      = ScenMean + tmpScenMean;
                    DSName        = "&DSName";
                    Scenario_Type = "&Scenario_Type";
                    if last then output;
    %end;
    %else %do;
                    ScenMean      = Mean;
                    DSName        = "&DSName";
                    Scenario_Type = "&Scenario_Type";
    %end;
            run;

            data _ScenMeans_;
                set _tmp_ScenMeans_ _ScenMeans_;
            end;
        %end;

        %end;
        
    data _ScenMeans_(keep= ScenMean DSName Scenario_Type);
        set _ScenMeans_;
        retain TotalScenMean 0;
        TotalScenMean = TotalScenMean + ScenMean;
        call symputX("TOTAL_SCEN_MEAN",TotalScenMean);
    run;

    proc sort data= _ScenMeans_;
        by ScenMean;
    run;

    data _ScenMeans_;
        set _ScenMeans_ end=Last;
        retain TotalScenCount 0;
        subScenCount = floor(ScenMean * &lookupSize /&TOTAL_SCEN_MEAN);
        TotalScenCount = TotalScenCount + subScenCOunt;

        if last then do;
            subScenCount = subScenCount + (&lookupSize - TotalScenCount);
            TotalScenCount = &lookupSize;
        end;
    run;

    %do nScen=1 %to &nRareEventScen;
        data _NULL_;    
            Index = &nScen;
            set _ScenMeans_ point=Index;
            call symput("SubScenCount",subScenCount);
            call symput("DSName",DSName);
            call symput("Scenario_Type", Scenario_Type);
            stop;
        run;
        
         /* Set the lookup table up */
        data _temp0_;
            do i=1 to &subScenCount;
                threshold=0;
                output;
            end;
        run;

            /*Case where the scenario is of Model type */
            %if 
                "&SCENARIO_TYPE" EQ "M" 
                %then %do;
            %HandleModelScenario(baseData=_temp0_, counter = &nScen); 
            %if "&abort"="1" %then %GOTO DONE_LOOKUP_RARE;
            %end;
        /* Case where the scenario is of type Bucketed */
            %else %do;
            %HandleBucketScenario(baseData=_temp0_, counter = &nScen);
            %if "&abort"="1" %then %GOTO DONE_LOOKUP_RARE;
        %end;
        
        /* Create the lookup table for this mixing interval  */
        /* and merge with other mixing intervals             */
        %if "&nScen" = "1" %then %do;
            data _lookups_ (keep= quantile); /*ecdf );*/
                set _lookup&nScen(rename=( Quant&nScen=Quantile));
            run;
        %end;
        %else %do;
            data _lookups_(keep=  quantile); /*ecdf );*/
                set _lookups_ _lookup&nScen(rename=( Quant&nScen=Quantile));
            run;
        %end;
        %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
            %put ERROR: Unable to create lookup table for mixing interval &nInt.;
            %let abort=1;
            %let reason=Unable to create lookup table for mixing interval &nInt;
            %goto DONE_LOOKUP_RARE;    
        %end;
        %end;
    
    proc sort data=_lookups_;
        by quantile;
    run;

	%if "&SCENARIO_TYPE" EQ "M" %then %do;
    	%CompactifyLookupArray( inData   = _lookups_,
                            quantVar = quantile,
                            outData  = _LookUp_Table_Rare
                            );
	%end;
	%else %do;
		data _LookUp_Table_Rare;
			set _lookups_ nobs=nobs;
			ecdf = 1-(1-_N_/(nobs+1))**2.5;
		run;
	%end;

    %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put ERROR: Unable to write out the lookup table as an array.;
        %let abort=1;
        %let reason=Unable to write out the lookup table as an array.;
        %goto DONE_LOOKUP_RARE;    
    %end;


    /* Try this SQL to create the lookup table macro */
    proc SQL noprint;
        SELECT 
            quantile
        INTO
            :RARE_LOOKUP_QUANTILES SEPARATED by " "
        FROM
            _LookUp_Table_Rare
        ;
    quit;


%DONE_LOOKUP_RARE:
%mend CreateRareEventLookupTable;

/* This macro takes a dataset name "inData" and a quantile  */
/* variable name "quantVar" assuming that those quantiles   */
/* were equiprobable and transforms the list of quantiles   */
/* into one that is distributed according to the importance */
/* sampling weights used.                                   */
%macro CompactifyLookupArray(inData   = ,
                             quantVar = , 
                             outData  =
                             );

    %let xyz=%sysevalf(1-(1-1/(&lookupSize +1))**2.5);
    data &outData(keep= ecdf &quantVar);
        format &quantVar BEST12.;
        set &inData(rename=(&quantVar=oldquantile)) nobs=ObsCount;
        retain      
                lastQuantile 0 
                lastEcdf 0
                Ecdf  &xyz
                newCount 1
                ;

        newEcdf = _N_/(obscount + 1);

        if newEcdf < Ecdf then do;
            lastQuantile = oldquantile;
            lastEcdf = newEcdf;
        end;
        else do;
            newCount = newCount + 1;
            &quantVar = ((Ecdf-lastEcdf)*oldquantile + (newEcdf-ecdf)*lastQuantile)/(newEcdf-lastEcdf);
            output;
            ecdf=1-(1-newCOunt/(&lookupsize+1))**2.5;
        end;

        if _N_=Obscount then do;
            do while( newCount<= &lookupsize );
                ecdf=1-(1-newCOunt/(&lookupsize+1))**2.5;
                newCount=newCOunt +1;
                &quantVar=oldquantile;
                output;
            end;
        end;
    run;
%mend;


/* This macro compactifies the severity priority ranges */
/* i.e. it combines two or more consecutuve ranges if   */
/* they have the same DATASET_NAME.                     */
%macro CompactifyRanges(data=, out=, ccuid=);
    data &out; 
        set &data; 
        where CCUID=&ccuid; 
    run;

    proc sort data=&out; 
        by LOWER_BOUND; 
    run;

    data &out; set &out;
      length dsname $32;
      retain dsname "";
      if dsname ne DATASET_NAME then output;
      dsname = DATASET_NAME;
      drop dsname;
    run;

    proc sort data=&out; 
        by descending LOWER_BOUND; 
    run;

    data &out; set &out;
      retain lbound -1;
      if lbound ne -1 then UPPER_BOUND = lbound;
      else UPPER_BOUND = 1;
      lbound = LOWER_BOUND;
      drop lbound;
    run;

    proc sort data=&out; 
        by LOWER_BOUND; 
    run;
%mend CompactifyRanges;
