/*************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *    NAME: orva_loss_sim_stream.sas                                             
 * PURPOSE: sub-macro of %orva_loss_sim() called once for each stream.            
 * SUPPORT: jariou
 *
 * PARAMETERS:
 *    startrep    = 1st rep number.
 *    nreps       = Number of reps for stream.
 *    streamNum   = Current stream number.
 *    seed        = Initial seed for stream.
 *
 * NOTES: Please see the accompanying doc for %orva_loss_sim().
 *        The only output of this macro are the datasets:
 *           _osimNaked_&streamNum 
 *           _osimInsured_&streamNum
 *
 * SUBORDINATE MACROS:
 *    %ApplyInsurance(Level=);                                               
 *    %RunSimulation(lookupTable=, applyKris=);
 *    %ApplyPlainInsurance(Level=);
 *    %ApplyBaselInsurance(Level=);
 *    %SimulateRareCount(rareCount=);
 **************************************************************************************/
%macro orva_loss_sim_stream(
                            startRep  =, 
                            nReps     =,
                            streamNum =,
                            seed      =
                            );

    %orvd_pois;
    data _osimNaked_&streamNum (keep= _rep_ loss)
        _osimInsured_&streamNum 
            %if "&coverageData1" ne "" and &CoverageCount^=0 %then %do;
                (Keep= _rep_ loss &InsuranceArray)
            %end;
            %else %do;
                (keep= _rep_ loss nloss)
            %end;
            ;

        /* Initialize the random stream */
        call streaminit(&seed);

        /* Always create a main lookup table array */
        array mainLookupTable(&lookupSize) _TEMPORARY_ (&LOOKUP_QUANTILES);

        /* If there is at least one rare event scenario, create  */ 
        /* the lookup table arrray for rare event simulation     */
        %if "&nRareEventScen" NE "0" %then %do;
            array rareLookupTable(&lookupSize) _TEMPORARY_ (&RARE_LOOKUP_QUANTILES);
        %end;


         %if "&coverageData1" ne "" and &CoverageCount^=0 %then %do;
                array Insurance_[&InsuranceContractCount] &InsuranceArray;
                array arInsurance_UID_1_[&CoverageCount] _TEMPORARY_ ( &Insurance_UID_1 );      
                array arInsurance_Index_1_[&CoverageCount] _TEMPORARY_ ( &Insurance_Index_1 );      
                array arDeductible_1_[&CoverageCount] _TEMPORARY_ ( &Deductible_1 );      
                array arLimit_1_[&CoverageCount] _TEMPORARY_ (&Limit_1 );                      
                array arEff_Agg_Limit_1_[&CoverageCount] _TEMPORARY_ (&Eff_Agg_Limit_1 );     
                array arPayout_Haircut_1_[&CoverageCount] _TEMPORARY_ (&Payout_Haircut_1 );    
                array arPolicyExpiryDate_1_[&CoverageCount] _TEMPORARY_ (&PolicyExpiryDate_1 );    
                array arInsurance_UID_2_[&CoverageCount] _TEMPORARY_ ( &Insurance_UID_2 );      
                array arInsurance_Index_2_[&CoverageCount] _TEMPORARY_ ( &Insurance_Index_2 );      
                array arDeductible_2_[&CoverageCount] _TEMPORARY_ (&Deductible_2 );            
                array arLimit_2_[&CoverageCount] _TEMPORARY_ (&Limit_2 );                      
                array arEff_Agg_Limit_2_[&CoverageCount] _TEMPORARY_ (&Eff_Agg_Limit_2 );              
                array arPayout_Haircut_2_[&CoverageCount] _TEMPORARY_ (&Payout_Haircut_2 );    
                array arPolicyExpiryDate_2_[&CoverageCount] _TEMPORARY_ (&PolicyExpiryDate_2 );    
        %end;


        /*  Later, I will set loss to sumNaked, then output to outsimNaked, 
            then set loss to sumInsured and output to outsimInsured.    */


            /* TODO: */ 
            /* Could replace the set _loss_ point=??? */
            /* by an array written to a file before   */
            /* slope = (lookupTable[&LookupSize] - lookupTable[&LookupSize-1]); */
 
        /* Loop through replications */
        do curr_rep = &startRep to &startRep + &nreps -1 ;
            set _loss_ point=curr_rep;

            %if "&coverageData1" ne "" and &CoverageCount^=0 %then %do;
                do ii =1 to &InsuranceContractCount;
                    Insurance_[ii]=0;
                end;
            %end;    
            /*- simulate severities -*/
            sumNaked =0;
            sumInsured=0;
            sumInsurance=0;

            /* Loop through individual losses */
            do i = 1 to nloss; 
                %if "&AVERAGE_KRI_EFFECT" NE "1" %then %do;
                    %RunSimulation( lookupTable = mainLookupTable, applyKris = Yes, rareEventScen=No);
                %end;
                %else %do;
                    %RunSimulation( lookupTable = mainLookupTable, applyKris = No, rareEventScen=No);
                %end;
            end; /* End of Loop through individual losses */
                    
            %if "&nRareEventScen" NE "0" %then %do;
                FParam_1 = &TOTAL_SCEN_MEAN ;
                %SimulateRareCount( rareCount = theRareCount);
                
                do j=1 to theRareCount;
                    %RunSimulation( lookupTable = rareLookupTable, applyKris = No, rareEventScen=Yes);
                    end;
            %end;

            _rep_  = curr_rep;
            loss   = sumNaked;                  
            output _osimNaked_&streamNum;

            loss   = sumInsured;    
            output _osimInsured_&streamNum;
    
        end; /* End of Loop through replications */
        stop;  /* for having used "point" */
   run;
                        
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
      %put ERROR: Unable to simulate aggregate loss (stream=&streamNum).;
      data _osimNaked_&streamNum; set _NULL_; run;
      data _ostatNaked_&streamNum; set _NULL_; run;
   %end;
   
%mend orva_loss_sim_stream;

%macro RunSimulation(lookupTable=, applyKris=, rareEventScen=);
 
                    LossIndex = (&LookupSize+1)*(1-rand("UNIFORM")**.4);
                    FL_Index    = Floor(LossIndex);
                    if FL_Index EQ 0 then do;
                        iLoss = (1-(1-LossIndex)**2.5)*&lookupTable[1];
                    end;
                    else if FL_Index GE &LookupSize then do;
                        iLoss = &lookupTable[&LookupSize]+(LossIndex-&LookupSize)*(&lookupTable[&LookupSize]-&lookupTable[&LookupSize-1]);      
/*                        iLoss = &lookupTable[&LookupSize];      */
                    end;
                    else do;
                        iLoss = &lookupTable[FL_Index]+(LossIndex-FL_Index)*(&lookupTable[FL_Index+1]-&lookupTable[FL_Index]);                        
/*                        iLoss = &lookupTable[FL_Index]+(1-(1-(LossIndex-FL_Index))**2.5)*(&lookupTable[FL_Index+1]-&lookupTable[FL_Index]);                        */
                            end;

                    /* TODO: */
                    /* Speed up can be accomplished here by putting all of this in an array */
                    /* Scale appropriately if KRI's are used */
                    %if "&applyKris" EQ "Yes" %then %do;
	                    /* Scale appropriately if KRI's are used */
	                    uChooseKriScale=1+int(&KRI_CHOICES*rand("UNIFORM"));
	                    set _tmp_kri_ point=uChooseKriScale;
	                    iLoss=iLoss * KRI_SCALE_EFFECT/&AVERAGE_KRI_EFFECT;
                    %end;

		            /* Added so that maximum loss does not escape to infinity */
		            /* but yet does not just truncate at MaxLoss */
                    %if "&rareEventScen" EQ "No" %then %do;
		            if iLoss > &MaxLoss then do;
                           iLoss = &MaxLoss;
		            end;
                    %end;

                    sumNaked=sumNaked+iloss;

                    /*- apply insurance contract -*/
                    %if "&coverageData1" ne "" and &CoverageCount^=0 %then %do;  
                        benefit_1 = 0;                    
                        benefit_2 = 0;                    
                        CoverageIndex = rand("TABLE", &pInsurance);

                        /* In case the probabilities of coverage sum to less than 1 */
                        if CoverageIndex LE &pInsuranceCount then do; 
                        /* Check if there is primary Insurance coverage */
                        if arInsurance_UID_1_[CoverageIndex] NE 0 then do; 
                                /*****************************************************************/
                                /* Choose which type of insurance coverage you want to apply     */
                                /*   Plain                                                       */
                                /*   Basel : Includes haircut effect and date of expiry effect   */
                                /*                                                               *
                                /*  Leave the one you want uncommented and comment the other one */
                                /*****************************************************************/
                                /* %ApplyPlainInsurance(Level=1);                                */
                                %ApplyBaselInsurance(Level=1);
                                /* Check if there is secondary Insurance coverage */
	                            if arInsurance_UID_2_[CoverageIndex] NE 0 then do;
                                   /* Choose type of insurance coverage  */
                                   /* See comments ~10 lines above       */
                                   /* %ApplyPlainInsurance(Level=2);     */
                                   %ApplyBaselInsurance(Level=2);    
	                        end;    
	                        iLoss = iLoss - benefit_1 - benefit_2;   

                            end; /* End of: Check if there is primary Insurance coverage */

                        end; /* End of: In case the probabilities of coverage sum to less than 1 */
                    %end; /* End of application of Insurance */

                    sumInsured = sumInsured+iloss;                
%mend;/* end of %RunSimulation */


%macro ApplyPlainInsurance(Level=);
    /*- handle aggregate limit when applicable -*/
    if arEff_Agg_Limit_&Level._[CoverageIndex] then do;  
        /*- if the accumulated benefits have not reached the aggregate limit, -*/
        /*- calculate the amount of insurance benefit -*/
        if sumInsurance < arEff_Agg_Limit_&Level._[CoverageIndex] then do; 
            benefit_&Level = max(0, min(iloss-arDeductible_&Level._[CoverageIndex],arLimit_&Level._[CoverageIndex]-arDeductible_&Level._[CoverageIndex]));
            sumInsurance=sumInsurance + benefit_&Level.;
            /*- check whether we passed the aggregate limit  -*/
            /*- this test will be true at most once  -*/
            if sumInsurance >= arEff_Agg_Limit_&Level._[CoverageIndex] then do; /*Level 3*/
                benefit_&Level. = benefit_&Level. + arEff_Agg_Limit_&Level._[CoverageIndex] -sumInsurance;
                sumInsurance = arEff_Agg_Limit_&Level._[CoverageIndex];   
            end;/*- we have reached the aggregate limit, no more insurance benefit -*/
        end;    
    end; 
    else do;
        benefit_&Level. = max(0,min(iloss-arDeductible_&Level._[CoverageIndex],arLimit_&Level._[CoverageIndex]-arDeductible_&Level._[CoverageIndex]));
    end; 

    Insurance_[arInsurance_Index_&Level._[CoverageIndex]] = benefit_&Level. + Insurance_[arInsurance_Index_&Level._[CoverageIndex]] ;
%mend; /* ApplyPlainInsurance */

%macro ApplyBaselInsurance(Level=);
    /* Apply insurance only if the policy is not expiring soon */
    /* The calculation was done in reverse before */
    if arPolicyExpiryDate_&Level._[CoverageIndex]-&simDate GE 90 then do;
    /*- handle aggregate limit when applicable -*/
    if arEff_Agg_Limit_&Level._[CoverageIndex] then do;  /*Level 1*/
        /*- if the accumulated benefits have not reached the aggregate limit, -*/
        /*- calculate the amount of insurance benefit -*/
        if sumInsurance < arEff_Agg_Limit_&Level._[CoverageIndex] then do; /*Level 2*/
                /*Just in case some Haircut is missing in the Insurance data */
                if arPayout_Haircut_&Level._[CoverageIndex] NE . then do;
                    benefit_&Level = arPayout_Haircut_&Level._[CoverageIndex] * max(0, min(iloss-arDeductible_&Level._[CoverageIndex],arLimit_&Level._[CoverageIndex]-arDeductible_&Level._[CoverageIndex]));
                end;
                else do;
            benefit_&Level = max(0, min(iloss-arDeductible_&Level._[CoverageIndex],arLimit_&Level._[CoverageIndex]-arDeductible_&Level._[CoverageIndex]));
                end;
            sumInsurance=sumInsurance + benefit_&Level.;
            /*- check whether we passed the aggregate limit  -*/
            /*- this test will be true at most once  -*/
            if sumInsurance >= arEff_Agg_Limit_&Level._[CoverageIndex] then do; /*Level 3*/
                benefit_&Level. = benefit_&Level. + arEff_Agg_Limit_&Level._[CoverageIndex] -sumInsurance;
                sumInsurance = arEff_Agg_Limit_&Level._[CoverageIndex];   
            end;/*- we have reached the aggregate limit, no more insurance benefit -*/
        end;    /*Level 2*/
    end; /*Level 1*/
    else do;
            benefit_&Level. = arPayout_Haircut_&Level._[CoverageIndex] * max(0,min(iloss-arDeductible_&Level._[CoverageIndex],arLimit_&Level._[CoverageIndex]-arDeductible_&Level._[CoverageIndex]));
    end; /*Level 1*/

    Insurance_[arInsurance_Index_&Level._[CoverageIndex]] = benefit_&Level. + Insurance_[arInsurance_Index_&Level._[CoverageIndex]] ;
    end;
%mend; /* ApplyBaselInsurance */

/* Fix for real  */
%macro SimulateRareCount(rareCount=);
        %SimPOIS(&rareCount);
%mend;
