/*****************************************************************************************
 *  Copyright (C), 2005
 *  SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    orva_range_merge.sas                                             
 *  TITLE:   Merges the Loss dataset with range data set and produces exposure size
 *  SUPPORT: sasdje
 *
 *  PARAMETERS:
 *
 *  data  =  the name of the loss dataset with the following variables:
 *             &CCV_NAMES 
 *
 *  dataOut = the name of the output dataset. Adds the variable exposure_size to the losses
 *
 *  missingExposureOut = 
 *
 *  missing_range = name of the missing ranges dataset with the following variables:
 *            &CCV_NAMES, START_MISSING_DATE, END_MISSING_DATE.
 *  
 *  global_range = name of the global range dataset with the following variables:
 *            &CCV_NAMES, START__DATE, END_DATE. 
 * 
 *  CCV_NAMES = list of cross class names used in the project usually:
 *           INTERNAL_BL_UID and INTERNAL_RC_UID                
 *  
 *  ostat = name of the out status dataset.  Will contain the following:
 *           status = 0 on success, 1 otherwise.
 *           reason = printable message.
 *
 *  interval = time interval size for which the analysis needs to be done, defaults to YEAR.    
 *             possibles values: YEAR, SEMIYEAR, QTR, MONTH2, MONTH, SEMIMONTH, WEEK2, WEEK
 *
 *      
 *****************************************************************************************/
 %macro orva_range_merge( data                = ,
                          dataOut             = loss_exposure,
						  missingExposureOut  = missing_exposure,
                          missing_range       = ,
                          global_range        = ,
                          CCV_NAMES           = INTERNAL_BL_UID INTERNAL_RC_UID , 
                          ostat               = status,
						  interval            = QTR);

/*%orvh_displaypaths;*/

    /*- create the status dataset upfront -*/
    %let status=0;
    %let reason=Successful;
    %let abort=0;
    data &ostat;
            format status 8. reason $100.;
            status=1;
            reason="Unknown";
    run;

    /*- remove anything in other output datasets up front -*/
    data &missingExposureOut; 
        set _NULL_; 
    run;
    
	/*- DO NOT do this for &dataOut, because the input &data is often the output &dataOut, 
	    so we lose the input if we clear &dataOut!!! -*/


	/*- check input parameters -*/
    %if &global_range=%str() %then %do;
        %put ERROR: The global_range dataset must be specified.;
        %let abort=1;
        %let reason=The global_range dataset must be specified.;
        %goto DONE;
    %end;
    %if &data=%str() %then %do;
        %put ERROR: The parameter data= must be specified.;
        %let abort=1;
        %let reason=Missing data= parameter;
        %goto DONE;
    %end;

    /*- get dates from global_range table -*/
    data _null_; set &global_range; 
       call symput( "range", end_date-start_date);
	   call symput( "global_end", end_date);
	   call symput( "global_start", start_date);
       call symput("number_of_intervals", intck("&interval", start_date, end_date));
    run;
    %if %eval(&range) <= 0 %then %do;
        %put ERROR: The start_date in the global range table is not less than the end_date;
        %let abort=1;
        %let reason=The start_date in the global range table is not less than the end_date;
        %goto DONE;
    %end;
 
    /*- handle exposures when there are no missing ranges -*/
    data _NULL_; set &global_range;
           n1 = intck("&interval", 0, start_date);
           diff_days = intnx("&interval",0, n1+1) - start_date;
           diff = intnx("&interval",0, n1+1) - intnx("&interval",0, n1);
           exposure_s = diff_days/diff;
           call symput("cutoff_s", intnx("&interval",0, n1+1)); 
           n2 = intck("&interval", 0, end_date);
           diff_days = end_date - intnx("&interval",0, n2);
           diff = intnx("&interval",0, n2+1) - intnx("&interval",0, n2);
           exposure_e =diff_days/diff;
           call symput("cutoff_e", intnx("&interval",0, n2)); 
	   call symput( "exposure_s", exposure_s);
	   call symput( "exposure_e", exposure_e);
    run;

    /* Counts divided by the sum of the exposures -----------------------*/
    %if &missing_range=%str() %then %do;
	    /*- if there are no missing ranges, exposure on middle pieces is 1 -*/
           /*- On the ends, exposure depend on global range ------------------*/
        %no_missing_ranges;
        %goto DONE;
    %end;
    %else %do;
        
		%let nobs=0;
		%let dsid=%sysfunc(open(&missing_range,I));
		%if &dsid NE 0 %then %do;
			%let nobs=%sysfunc(attrn(&dsid,nlobs));
			%let rc=%sysfunc(close(&dsid));
		%end;
		%if &nobs eq 0 %then %do;	
			%no_missing_ranges;
 			%goto DONE;
		%end;	
	%end;
    
    
    /*- compute exposures -------------------------------------*/

    data _exposure; set &missing_range;
        exposure_size = (end_missing_date - start_missing_date) / &range;
        if exposure_size LE 0 then do;
           put "ERROR: Observation " _N_ " of the missing range table contains invalid dates.";
           put Start_Missing_Date=;
           put End_Missing_Date=;
           call symput( "abort", 1); 
           call symput( "reason", "ERROR: Observation " || _N_ || " of the missing range table contains invalid dates." );        
        end;
    run;


    /*- sum the exposures to get the collection of holes -*/
    proc summary  data=_exposure;
          output out=_exposure(drop=_type_ _freq_) sum= exposure_size n=COUNT;
          var exposure_size; 
          by &CCV_NAMES;
    run;
    data _exposure; set _exposure;
       where count > 0;
	   if exposure_size LE 0 or exposure_size = . then do;
           put "ERROR: Observation with (&CCV_NAMES)= " &CCV_NAMES " in missing range table contains invalid dates." ;
           call symput( "abort", 1); 
           call symput( "reason", "Invalid dates in missing range table");
       end;
              /* We have sum of the holes need sum of collection parts */
       exposure_size =1 - exposure_size;
       if exposure_size < 0 or exposure_size = . then do;
           put "ERROR: Observation with (&CCV_NAMES)= " &CCV_NAMES " in missing range table contains invalid dates." ;
           put "ERROR: The sum of the missing dates is greater than the global missing range";
           call symput( "abort", 1); 
           call symput( "reason", "The sum of the missing dates is greater than the global missing range");
       end;
       keep &CCV_NAMES exposure_size;
    run;
     
   
         
    /*- merge with loss data missing_ranges  -*/
    
            /* This produces a data set with a interval indicator variable    */
            /* This allows us to merge with the exposures                     */
    data _sortedData; set &data;
         interval_indicator = intck("&interval", &Global_Start, FINANCIALDATE);
		 in_brkdwni = 1;  /* add a flag to help us later (during merge) identify BLUIDs that have no losses */
    run;

    proc sort data=_sorteddata out=_sortedData; 
       by INTERNAL_BL_UID INTERNAL_RC_UID PROCESS_UID CLASS_1-CLASS_10 interval_indicator;
    run;

    proc sort data=&missing_range out=_exposure; 
       by INTERNAL_BL_UID INTERNAL_RC_UID PROCESS_UID CLASS_1-CLASS_10;
    run;

           /* Now expand out the missing ranges to the same set of intervals */
           /* and compute exposures for those intervals ---------------------*/
    data _exposure (compress=binary); set &missing_range;
         by INTERNAL_BL_UID INTERNAL_RC_UID PROCESS_UID CLASS_1-CLASS_10;
         retain first_time 1;
         array dates[&number_of_intervals];
         array e[&number_of_intervals];

            /* Initialize the date array once -----------*/
         if  first_time = 1 then do;
             first_time = 0;
             
             dates[1] = &Global_Start;
             dates[&number_of_intervals] = &Global_End;
             do i = 2 to &number_of_intervals-1;
                dates[i] = intnx("&interval", &Global_Start, i-1); 
             end;
         end;

           /* Initialize the exposures for each ccv ------*/
         if first.CLASS_10 then do;
            do i = 1 to &number_of_intervals; 
               e[i] = 1;
            end;
         end;
         
         start_idx = intck("&interval", &Global_start, start_missing_date)+1;
         end_idx = intck("&interval", &Global_start, end_missing_date)+1;
         do i = start_idx to end_idx;
                /* Three cases we are at the begining middle or end of a missing range */
             if i = start_idx then do;
                e[i] = e[i] - (dates[i+1]-start_missing_date)/ (dates[i+1] - dates[i]);
             end;
             else 
             if i = end_idx then do;
                e[i] = e[i] - (end_missing_date - dates[i])/ (dates[i+1] - dates[i]);
             end;
             else e[i] = 0;
             if e[i] < 0 then e[i] = 0;
         end;

            /* output the exposures ----------------------------*/
         if last.CLASS_10 then do;
            do i = 1 to &number_of_intervals;
                exposure_size2 = e[i];
                interval_indicator = i-1;
                financialdate=dates[i];
				if exposure_size2 ^=1 then
                   output;
            end;
         end;
         keep INTERNAL_BL_UID INTERNAL_RC_UID PROCESS_UID CLASS_1-CLASS_10 
             exposure_size2 interval_indicator CCUID CCVUID_0 CCVUID_1 financialdate;
    run;

         /* Merge the exposure file with the output dataset -----------------*/
    proc sort data=_exposure; 
       by INTERNAL_BL_UID INTERNAL_RC_UID PROCESS_UID CLASS_1-CLASS_10 interval_indicator;
    run;

    data &dataOut (compress=binary);
		merge _sortedData _exposure; 
        by INTERNAL_BL_UID INTERNAL_RC_UID PROCESS_UID CLASS_1-CLASS_10 interval_indicator;
        if exposure_size = . then exposure_size = 1;
		if exposure_size2 ^=. then exposure_size = exposure_size * exposure_size2;
              /* get rid of exposures from lower level bls ---------------------*/
        if exposure_size = 1 and FSCALE = . then delete;
		if missing(in_brkdwni) then delete; /* Drop obs from _exposure that don't correspond to BLUIDs in 
		brkdwni (the loss dataset); That is, don't keep info about missing ranges for BLUIDs that had no losses.*/
        if FSCALE = . then FSCALE = 0;

               /* Handle partial month/year due to start/end date ------------------*/
        if FINANCIALDATE <= &cutoff_s then exposure_size = &exposure_s * exposure_size;
        else
        if FINANCIALDATE >= &cutoff_e then exposure_size = &exposure_e * exposure_size;
		drop exposure_size2;
    run;


     /* add two dummy obs for exposure at the ends ----------------------*/
    data &dataOut (compress=binary); 
		set &dataout; 
        by INTERNAL_BL_UID INTERNAL_RC_UID PROCESS_UID CLASS_1-CLASS_10 interval_indicator;
        retain f l;
        
        if first.interval_indicator then do;
           f = 0;
           l = 0;
        end;
           
             /* If there is already and observation at the begining and end don't add one */
        if FINANCIALDATE <= &cutoff_s then f = 1;
        else
        if FINANCIALDATE >= &cutoff_e then l = 1;

        output;
        if first.interval_indicator then do;
           if &exposure_s < 1  and f = 0 then do;
              FINANCIALDATE = &GLOBAL_START;
              FSCALE = 0;
              AMOUNT = .;
              exposure_size = &exposure_s;
              output;
           end;
           if &exposure_e < 1 and l = 0 then do;
              FINANCIALDATE = &GLOBAL_END;
              FSCALE = 0;
              AMOUNT = .;
              exposure_size = &exposure_e;
              output;   
           end;
        end;

        drop f l;
        
    run;


    /*- Produce the exposure needed for frequency fitting, in %orva_freq_var_select() -*/
	data &missingExposureOut(compress=binary 
	                         keep=CCUID internal_bl_uid start_missing_date end_missing_date 
								  new_month global_start global_end exposure);
		set &missing_range;
		format new_month global_start global_end date9.;
		global_start = &global_start;
		global_end = &global_end;
		j=intck("month",intnx("month",start_missing_date,0), intnx("month",end_missing_date,1));
		do i=1 to j;
			new_month = intnx("month", start_missing_date,i-1);
			if i=1 then do;
				start_of_curr_interval = intnx("&interval", start_missing_date, 0);
				end_of_curr_interval = intnx("&interval", start_missing_date, 1);
				DaysInCurrInterval = end_of_curr_interval - start_of_curr_interval;
				exposure = (day(start_missing_date)-1) / DaysInCurrInterval;
			end;
			else if i=j then do;
				DaysLeftInMonth = intck("day", end_missing_date, intnx("month", end_missing_date, 1))-1;
				start_of_curr_interval = intnx("&interval", end_missing_date, 0);
				end_of_curr_interval = intnx("&interval", end_missing_date, 1);
				DaysInCurrInterval = end_of_curr_interval - start_of_curr_interval;
				exposure = DaysLeftInMonth / DaysInCurrInterval;
			end;
			else exposure = 0;
			output;
		end;
	run;

   
%DONE:
    %if &abort %then %do;
	    %put ERROR: orva_krimerge() macro ended abnormally.;
	    data &ostat;
	        format status 8. reason $100.;
	        status=1;
	        reason="&reason";
	    run;
    %end;    
    %else %do;
	    data &ostat;
	        format status 8. reason $100.;
	        status=0;
	        reason="Success";
	    run;
    %end;
    
    /*- delete unwanted data sets -*/
    proc datasets library=work memtype=data nolist nodetails;
       delete  _exposure _tmp_ _sortedData _ccv;
    run; quit;

%mend;



%macro no_missing_ranges;

           
    data &dataOut (compress=binary); set &data;
       if FINANCIALDATE <= &cutoff_s then exposure_size = &exposure_s;
       else
       if FINANCIALDATE >= &cutoff_e then exposure_size = &exposure_e;
       else
       exposure_size = 1;
    run;

	/*- output global_start and global_end dates for frequency fitting -*/
	data &missingExposureOut;
		format global_start global_end date9.;
		global_start = &global_start;
		global_end = &global_end;
		internal_bl_uid = -1;
	run;

%mend no_missing_ranges;
