/******************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_WBL.sas
 *   
 *   -Provides support for the Weibull distribution-
 *
 *	This file contains macro fragments to implement calculation specific to the 
 *	Weibull distribution
 *
 *  The macro %ORVD_WBL is at the end of this file. It is an empty macro that takes
 *  no parameter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_WBL macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 *  The list of other macro files that rely on fragments in this file:
 *
 *   File                    Fragments it depends on (Approx Line # where fragement is called)
 *   orva_cdfplot.sas        CdfWBL(170)
 *   orva_gfit.sas           CdfWBL(188)
 *   orva_histplot.sas       CdfWBL(176), PdfWBL(177)
 *   orva_los_sim.sas        SdfWBL(263), SimWBL(408)
 *   orva_probplot.sas       CdfWBL(160)
 *   orva_quantplot.sas         QuantWBL(118)
 *   orva_sfitall.sas        FitWBL(304), MakeLLGoodWBL(320), NoMissingMOMWBL(379)
 *   orva_sfitall_model.sas  ModelFitWBL(287), MomentsWBL(391)
 *   orva_sfitall_mom.sas    MomFitWBL(78)
 *
 ******************************************************************************/

/*****************************************************************************************
 *   Needed for: %orva_cdfplot     ~line 170
 *               %orva_gfit        ~line 188
 *               %orva_histplot    ~line 176
 *               %orva_probplot    ~line 160
 *
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        returnVal = name of variable inton which result is stored
 *
 *****************************************************************************************/
%macro CdfWBL(x,returnVal);
       &returnVal=1-exp(-(&x/&Param_1)**&Param_2);    
%mend CdfWBL;


/*****************************************************************************************
 *   Needed for: %orva_loss_sim    ~line 263
 *
 *   What it does  :  Evaluate the survival distribution function (SDF)
 *
 *   Parameters    :
 *              x     = value at which the SDF is evaluated
 *        returnVal = name of variable inton which result is stored
 *
 *****************************************************************************************/
%macro SdfWBL(x,returnVal);
       &returnVal=exp(-(&x/&Param_1)**&Param_2);    
%mend SdfWBL;

/*****************************************************************************************
 *   Needed for: %orva_histplot    ~line 176
 *
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        returnVal = name of variable inton which result is stored
 *
 *****************************************************************************************/
%macro PdfWBL(x,returnVal);
       &returnVal = &Param_2 /&x * (&x/&Param_1)**&Param_2 * exp(-(&x/&Param_1)**&Param_2);    
%mend PdfWBL;

/*****************************************************************************************
 *   Needed for: %orva_quantplot    ~line 118
 *
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *
 *****************************************************************************************/
%macro QuantWBL;
       data _temp_;
          set _temp_ nobs=nobs end=eof;
		retain truncdf;

		if _N_ = 1 then do;
			if THRESHOLD = 0 then do;
				truncdf =0;
			end;
			else do;
				%CDFWBL(THRESHOLD, truncdf);
			end;
		end;

		cdf   = ecdf*(1-truncdf)+truncdf;
		quant = &Param_1 * (-log(1-cdf))**( 1 / &Param_2 );
     run;
%mend QuantWBL;

/*****************************************************************************************
 *   Needed for: %orva_loss_sim    ~line 408
 *
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *
 *****************************************************************************************/
%macro SimWBL(loss);
    &loss = rand("WEIBULL", Param_2, Param_1);
%mend SimWBL;

/*****************************************************************************************
 *   Needed for: %orva_sfitall    ~line 304
 *				 %orvr_sfit_stats	???
 *
 *   What it does  :  Get an initial value of the parameters
 *
 *   No Parameters    
 *
 *****************************************************************************************/
%macro InitFitWBL;
    %let nparms=2;    

	/*- initial guess based on percentiles -*/     
	data _null_;
		g=log(log(4))/log(log(4/3));
		sigma=exp((g*log(&sample_q1)-log(&sample_q3))/(g-1));
		c=log(log(4))/(log(&sample_q3)-log(sigma));
		call symput("Param_1_init",sigma);
		call symput("Param_2_init",c);
	run;

	%put ******************************;
	%put Distribution   =  &dist;
	%put Observations   =  &obsCount;
	%put Param_1_init    =  &Param_1_init;
	%put Param_2_init    =  &Param_2_init;
	%put ******************************;

%mend InitFitWBL;

/*****************************************************************************************
 *   Needed for: %orva_sfitall    ~line 304
 *
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *
 *****************************************************************************************/
%macro FitWBL;
	ods listing close;
    ods output 
		IterStop            =  _logl_
		ConvergenceStatus   =  _cs_ 
		covmat              =  _covmat_ 
		;	

	proc nlp 
		data   = _workdata_ 
		cov    = 4 pcov 
		maxit  = 10000 
		maxfu  = 10000 
		tech   = qn 
		outest = _parmnlp_
		;

		parms 
				Param_1=&Param_1_init, 
				Param_2=&Param_2_init;
		bounds 
				Param_1>0, 
				Param_2>0;

		max logf;

                        if &THRESHOLDPERCENT >= 0 then do;
			    lsdf = &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT)*
                                 logcdf("weibull",Threshold, Param_2, Param_1);
                        end; 
                        else
			if THRESHOLD=0 then do;
				lsdf=0;
			end;
			else do;
				lsdf=(THRESHOLD/(Param_1))**param_2;
			end;

			logf = 	lsdf + 
					log(Param_2) +
					(Param_2-1) * log(AMOUNT) - 
					Param_2*log(Param_1) -
					(AMOUNT/(Param_1))**Param_2
					;

	run;
	quit;

	ods listing;
%mend FitWBL;

/*****************************************************************************************
 *   Needed for: %orva_sfi_kri    ~line 304
 *
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *
 *****************************************************************************************/
%macro FitKriWBL;
	ods listing close;
    ods output 
		IterStop          =  _logl_
		ConvergenceStatus =  _cs_ 
		covmat            =  _covmat_ 
		;	

	proc nlp 
		data   = _LogWorkdata_KRI_ 
		cov    = 4 pcov 
		maxit  = 10000 
		maxfu  = 10000 
		tech   = qn 
		outest = _parmnlp_
		;

		parms 
				Param_1 = &Param_1_init, 
				Param_2 = &Param_2_init
				;

		bounds 
				Param_1>0, 
				Param_2>0
				;

		max logf;
                        if &THRESHOLDPERCENT >= 0 then do;
			    lsdf = &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT)*
                                logcdf("weibull",Threshold, Param_2, Param_1*KRI_SCALE);
                        end; 
                        else
			if THRESHOLD=0 then do;
				lsdf=0;
			end;
			else do;
				lsdf = (THRESHOLD / Param_1 / KRI_SCALE)**Param_2;
			end;

			logf = 	lsdf + 
					log(Param_2) + 
					(Param_2-1)*LogAMOUNT - 
					Param_2 * ( log(Param_1)+ KRI_LOG ) - 
					(AMOUNT/Param_1/KRI_SCALE)**Param_2
					;
	run;
	quit;

	ods listing;
%mend FitKriWBL;

/*****************************************************************************************
 *   Needed for: %orva_sfitall    ~line 320
 *
 *   What it does  :  Modify loglikelihood for cases when we had to take the log of the 
 *					  observations first. Here, it does nothing.
 *
 *   NO Parameters    
 *
 *****************************************************************************************/
%macro MakeLLGoodWBL;
    %let sumlog=0;
%mend MakeLLGoodWBL;
       
%macro MakeLLGoodKRIWBL;
    %let sumlog=0;
%mend MakeLLGoodKRIWBL;
       
/*****************************************************************************************
 *   Needed for: %orva_sfitall    ~line 379
 *
 *   What it does  :  Checks whether some moments are missing.
 *					  Here does nothing because all moments exist for Weibull distribution
 *
 *   No Parameters 
 *
 *****************************************************************************************/
%macro NoMissingMOMWBL(parmdata=);

%mend NoMissingMOMWBL;

/*****************************************************************************************
 *   Needed for: orva_sfitall_model.sas
 *								%FindMoments ~line 392
 *
 *   What it does  :  Calculates the moments
 *
 *   No Parameters 
 *
 *****************************************************************************************/
%macro MomentsWBL;
    %do i=1 %to 4;
        m&i=(param_1**&i)*exp(lgamma(1+&i/param_2));
    %end;
    mean=m1;
    std=sqrt(m2-m1*m1);
    skewness=(m3-3*m1*m2+2*(m1**3))/(std**3);
    kurtosis=(m4-4*m1*m3+6*(m1**2)*m2-3*(m1**4))/(std**4);    
%mend MomentsWBL;

%macro RealMomentsWBL;
	%MomentsWBL
%mend RealMomentsWBL;


/*****************************************************************************************
 *   Needed for: %orva_sfitall_mom.sas      ~line 78
 *
 *   What it does  :  Does method of moments estimation and assigns the corresponding 
 *                    values to the location, param_1, param_2, and shape2 macros
 *
 *   No Parameters 
 *
 *****************************************************************************************/
%macro MomFitWBL;
   data _temp_;
        m1      = &mean;
        m2      = &variance + &mean * &mean;
		Param_1 = m1;
        Param_2 = 1;
	    call symput("converge",m1*1e-8);
    run;
	Ods listing close;
	Ods output obsused = _cs_;
	proc model data=_temp_;
		dependent Param_2;
		eq.one = m1*sqrt(gamma(1+2/param_2))-sqrt(m2)*gamma(1+1/param_2);
        solve param_2/maxiter=10000 converge=&converge out=_sol_(keep=param_2);
	run;
	quit;
	Ods listing;
	data _null_;
		set _cs_;
		if label1 = "Failed" then do;
			test = 1;
			if cvalue1 = 1 then do;
		    		call symput("status",1);
				call symput("reason","Unable to compute MOM estimates");
			end;
		end;
	run;
	
	data _null_;
	    set _sol_;
        param_1=&mean/gamma(1+1/param_2);
	    call symput("param_1",param_1);
        call symput("param_2",param_2);
        run;		
%mend MomFitWBL;

/*****************************************************************************************
 *   Needed for: %orvr_sfitstats.sas      ~line 78
 *
 *   What it does  :  Displays the estimated parameters
 *
 *   No Parameters 
 *
 *****************************************************************************************/
%macro ParamsWBL;
	data foo (keep=Variable Estimate StdErr TValue probt);
		format Variable $12.;
		set &sparm;
		Variable="Scale";
		Estimate=param_1;
		StdErr=sqrt(cov_1);
		TValue=estimate/StdErr;
		probt=(1-probt(abs(tvalue),obscount -&KRI_ADJUST-&Param_count))*2;
		output;
		Variable="Shape1";
		Estimate=param_2;
		stderr=sqrt(cov_3);
		TValue=estimate/stderr;
		probt=(1-probt(abs(tvalue),obscount -&KRI_ADJUST-&Param_count))*2;
		output;
    run;

        
	data _NULL_;set foo;
          file print ods=(template='Stat.Reg.ParameterEstimates');
          put _ods_;
	run;
%mend ParamsWBL;

/*****************************************************************************************
 *   Needed for: %orvr_sfitstats.sas      ~line 78
 *
 *   What it does  :  Displays the estimated parameters
 *
 *   No Parameters 
 *
 *****************************************************************************************/
%macro CovmatWBL;
	data foo (keep=_name_ param_1 param_2 );
		format _name_ $12.;
		set &sparm;
		_name_="Scale";
		param_1=cov_1;
		param_2=cov_2;
		output;
		_name_="Shape1";
		param_1=cov_2;
		param_2=cov_3;
		output;
    run;

	proc print data=foo noobs label;
		var _name_ param_1 param_2 ;
		label _name_='00'x param_1="param_1" param_2="param_2" ;
	run;
%mend CovmatWBL;


/*****************************************************************************************
 *   What it does  :  Calculates the loglikelihood contribution for each observation
 *
 *   No Parameters 
 *
 *****************************************************************************************/
%macro LLWBL(var,retVal);
	if THRESHOLD=0 then lsdf=0;
	else lsdf=(THRESHOLD/param_1)**param_2;

	&retVal = lsdf+log(param_2)+(param_2-1)*log(AMOUNT)-param_2*log(param_1)-(AMOUNT/param_1)**param_2;
%mend LLWBL;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_WBL;
    %let EXTRA_PARAMS= ;
    %let NEEDED_INPUT= ;
    %let SCALE_TRANSFORM= ID;
    %let DIST_NAME= Weibull;
    %let PARAM_COUNT= 2;
	data WBL_Status;
		status = 0;
	run;
%mend ORVD_WBL;
