/*-------------------------------------------------------------------
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * %ORVR_FCMatrix(model_uid=, options=)
 *
 * Produces report for Frequency Correlation/Covariance Matrices
 *
 * Called by %orva_advanced() - not intended to be called directly!
 *
 * Input:
 *    model_uid - UID for model.
 *    options - specifies which parts of the report to produce, 
 *              separated by spaces: 
 *              Currently none supported
 *
 * Required tables (from RDBMS):
 *   varprj.MODEL
 *
 * Required tables (in project result set area):
 *   varprj.ccmapval_0, 
 *   varprj.ccmapval_1
 *   varprj.cc
 *   varprj.fcorr
 *
 *-------------------------------------------------------------------*/ 
%macro ORVR_FCMatrix(model_uid=,options=);	


  *** get correlation data;
  data _corr;
    set varprj.fcorr;
    where _TYPE_='CORR';
    format CCUID best.;
    CCUID = substr(_NAME_,6);
    drop _TYPE_ ;
  run;

  /***- do we have data? -***/
  %let nobs=0;
  %let dsid=%sysfunc(open(_corr,I));
  %if &dsid ^= 0 %then %do;
    %let nobs=%sysfunc(attrn(&dsid,nlobs));
    %let rc=%sysfunc(close(&dsid));
  %end;
  
  %if &nobs = 0 %then %do;
  	proc report data=_fake_ &reportSpec;	
		column fakevar;
		define fakevar / display 'Frequency Correlation Data is not available for project' ;
	run; quit;
	%goto DONE_CORR;
  %end;
  

  *** Adjust the column order to match the row order;
  proc sort data=_corr; by CCUID; run;
  data _null_; set _corr  end=last;
      length ordered_names $ 10000;
      retain ordered_names;
      ordered_names = trim(ordered_names)||' '||_NAME_;
      if last = 1 then call symput("Ordered_Names", ordered_names);
  run;

  *** get Names for Cross Classification Variables;
  proc sql noprint;
      create table corr as
      select * 
      from _corr w,
          varprj.cc cc,
          varprj.ccmapval_0 ccv0,
          varprj.ccmapval_1 ccv1
      where w.CCUID = cc.CCUID and
            cc.CCVUID_0 = ccv0.CCVUID_0 and 
            cc.CCVUID_1 = ccv1.CCVUID_1 order by CCUID;
  quit;
  data corr;
    set corr;
    drop CCUID CCVUID_0 CCVUID_1;
  run;

    /*- fix up cc strings -*/
    proc sort data=corr; by CCUID; run;
    data corr; set corr;
      retain foo;
      length foo $ 128;
      if trim(left(foo)) ne "" and trim(left(foo)) eq trim(left(CCVVALUE_0)) then do;
         foo = CCVVALUE_0;
         CCVVALUE_0="";
      end;
      else do; 
         foo=CCVVALUE_0; 
      end;
    run;

  *** run report;
  footnote1; %cleartitles; 
  title2 &titleSpec2 'Frequency Correlation Matrix';
  title3 &titleSpec3 'This report contains the Frequency Correlation Matrix used for VaR calculations';
  title4 &titleSpec4 "(Model=&MODEL_NAME Model Author=&FULL_NAME)";
  footnote1 &footnoteSpec1;
  proc report data= corr &reportSpec;
     column ('Classifications' CCVVALUE_0 CCVVALUE_1) _NAME_ &Ordered_Names;
	 define CCVVALUE_0 / left "&MAPNAME0" ;
	 define CCVVALUE_1 / display left "&MAPNAME1" ;
     define _NAME_/ ' ' Style=[font_weight=bold foreground=BLUE];
  run;
  
%DONE_CORR:

  *** get Covariance Data;
  data _cov;
    set varprj.fcorr;
    where _TYPE_='COV';
    format CCUID best.;
    CCUID = substr(_NAME_,6);
    drop _TYPE_;
  run;
    
  /***- do we have data? -***/
  %let nobs=0;
  %let dsid=%sysfunc(open(_cov,I));
  %if &dsid ^= 0 %then %do;
    %let nobs=%sysfunc(attrn(&dsid,nlobs));
    %let rc=%sysfunc(close(&dsid));
  %end;
  
  %if &nobs = 0 %then %do;
  	proc report data=_fake_ &reportSpec;	
		column fakevar;
		define fakevar / display 'Frequency Covariance Data is not available for project' ;
	run; quit;
	%goto DONE;
  %end;
  
  *** Adjust the column order to match the row order;
  proc sort data=_cov; by CCUID; run;
  data _null_; set _cov  end=last;
      length ordered_names $ 10000;
      retain ordered_names; 
      ordered_names = trim(ordered_names)||' '||_NAME_;
      if last = 1 then call symput("Ordered_Names", ordered_names);
  run;
 
  *** get Names for Cross Classification Variables;
  proc sql noprint;
      create table cov as
      select * 
      from _cov w,
          varprj.cc cc,
          varprj.ccmapval_0 ccv0,
          varprj.ccmapval_1 ccv1
      where w.CCUID = cc.CCUID and
            cc.CCVUID_0 = ccv0.CCVUID_0 and 
            cc.CCVUID_1 = ccv1.CCVUID_1 order by CCUID;
  quit;

    /*- fix up cc strings -*/
    proc sort data=cov by CCUID; run;
    data cov; set cov;
      retain foo;
      length foo $ 128;
      if trim(left(foo)) ne "" and trim(left(foo)) eq trim(left(CCVVALUE_0)) then do;
         foo = CCVVALUE_0;
         CCVVALUE_0="";
      end;
      else do; 
         foo=CCVVALUE_0; 
      end;
    run;

  *** run report;
  footnote1; %cleartitles; 
  title2 &titleSpec2 'Frequency Covariance Matrix';
  title3 &titleSpec3 'This report contains the Frequency Covariance Matrix used for VaR calculations';
  title4 &titleSpec4 "(Model=&MODEL_NAME Model Author=&FULL_NAME)";
  footnote1 &footnoteSpec1;
  proc report data= cov &reportSpec;
    column ('Classifications' CCVVALUE_0 CCVVALUE_1) _NAME_ &Ordered_Names;
	define CCVVALUE_0 / left "&MAPNAME0" ;
	define CCVVALUE_1 / display left "&MAPNAME1" ;
    define _NAME_/ '  ' Style=[font_weight=bold foreground=BLUE];				
  run;
  

%DONE:
 
  /*- delete unwanted data sets -*/
  proc datasets library=work memtype=data nolist nodetails;
    delete _cov cov _corr corr
  run; quit;



%mend ORVR_FCMatrix;
